# Copyright 2022 Cisco Systems, Inc. and its affiliates

import unittest
from unittest import TestCase, mock
from uuid import uuid4

from requests import Request

from catalystwan import USER_AGENT
from catalystwan.exceptions import CatalystwanException
from catalystwan.vmanage_auth import UnauthorizedAccessError, vManageAuth


class MockResponse:
    def __init__(self, status_code: int, text: str):
        self._status_code = status_code
        self._text = text
        self.request = Request()

    def cookies(self) -> str:  # TODO
        return "JSESSIONID=xyz"

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def text(self) -> str:  # TODO
        return self._text


def mock_request_j_security_check(*args, **kwargs):
    url_response = {
        "https://1.1.1.1:1111/j_security_check": {
            "admin": MockResponse(200, ""),
            "invalid_username": MockResponse(200, "<html>error</html>"),
        }
    }

    full_url = kwargs.get("url", "")
    data = kwargs.get("data", {})
    if full_url in url_response:
        return url_response[full_url][data["j_username"]]

    return MockResponse(404, "error")


def mock_valid_token(*args, **kw):
    return MockResponse(200, "valid-token")


def mock_invalid_token_status(*args, **kw):
    return MockResponse(503, "invalid-token")


def mock_invalid_token_format(*args, **kw):
    return MockResponse(200, "<html>error</html>")


class TestvManageAuth(TestCase):
    def setUp(self):
        self.base_url = "https://1.1.1.1:1111"
        self.password = str(uuid4())

    @mock.patch("requests.post", side_effect=mock_request_j_security_check)
    def test_get_cookie(self, mock_post):
        # Arrange
        username = "admin"
        security_payload = {
            "j_username": username,
            "j_password": self.password,
        }
        auth = vManageAuth(self.base_url, username, self.password)
        # Act
        auth.get_cookie()

        # Assert
        mock_post.assert_called_with(
            url="https://1.1.1.1:1111/j_security_check",
            data=security_payload,
            verify=False,
            headers={"Content-Type": "application/x-www-form-urlencoded", "User-Agent": USER_AGENT},
        )

    @mock.patch("requests.post", side_effect=mock_request_j_security_check)
    def test_get_cookie_invalid_username(self, mock_post):
        # Arrange
        username = "invalid_username"
        security_payload = {
            "j_username": username,
            "j_password": self.password,
        }
        auth = vManageAuth(self.base_url, username, self.password)
        # Act
        with self.assertRaises(UnauthorizedAccessError):
            auth.get_cookie()

        # Assert
        mock_post.assert_called_with(
            url="https://1.1.1.1:1111/j_security_check",
            data=security_payload,
            verify=False,
            headers={"Content-Type": "application/x-www-form-urlencoded", "User-Agent": USER_AGENT},
        )

    @mock.patch("requests.cookies.RequestsCookieJar")
    @mock.patch("requests.get", side_effect=mock_valid_token)
    def test_fetch_token(self, mock_get, cookies):
        # Arrange
        valid_url = "https://1.1.1.1:1111/dataservice/client/token"
        auth = vManageAuth(self.base_url, "admin", self.password)

        # Act
        token = auth.fetch_token(cookies)

        # Assert
        self.assertEqual(token, "valid-token")
        mock_get.assert_called_with(
            url=valid_url,
            verify=False,
            headers={"Content-Type": "application/json", "User-Agent": USER_AGENT},
            cookies=cookies,
        )

    @mock.patch("requests.cookies.RequestsCookieJar")
    @mock.patch("requests.get", side_effect=mock_invalid_token_status)
    def test_incorrect_xsrf_token_status(self, mock_get, cookies):
        auth = vManageAuth("http://invalid.response", "admin", self.password)
        with self.assertRaises(CatalystwanException):
            auth.fetch_token(cookies)

    @mock.patch("requests.cookies.RequestsCookieJar")
    @mock.patch("requests.get", side_effect=mock_invalid_token_format)
    def test_incorrect_xsrf_token_format(self, mock_get, cookies):
        auth = vManageAuth("http://invalid.response", "admin", self.password)
        with self.assertRaises(CatalystwanException):
            auth.fetch_token(cookies)


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 Cisco Systems, Inc. and its affiliates

# type: ignore
import unittest
from unittest.mock import MagicMock, patch

from attr.exceptions import NotAnAttrsClassError
from parameterized import parameterized  # type: ignore

from catalystwan.api.administration import (
    AdministrationSettingsAPI,
    ClusterManagementAPI,
    ResourceGroupsAPI,
    UserGroupsAPI,
    UsersAPI,
)
from catalystwan.dataclasses import (
    Certificate,
    CloudConnectorData,
    CloudServicesSettings,
    Connection,
    Organization,
    Password,
    ServiceConfigurationData,
    Vbond,
)
from catalystwan.endpoints.administration_user_and_group import (
    ResourceGroup,
    ResourceGroupSwitchRequest,
    ResourceGroupUpdateRequest,
    User,
    UserAuthType,
    UserGroup,
    UserGroupTask,
    UserResetRequest,
    UserRole,
    UserUpdateRequest,
)
from catalystwan.exceptions import CatalystwanException
from catalystwan.utils.certificate_status import ValidityPeriod
from catalystwan.utils.creation_tools import create_dataclass

password_dataclass = Password(old_password="old_password_123", new_password="new_password_123")
certificate_dataclass = Certificate("cert", "Name", "Surname", "name@sur.name", ValidityPeriod.ONE_YEAR, 10)
vbond_dataclass = Vbond("1.1.1.1", 1234)
organization_dataclass = Organization("My org name", 1)


class TestUsersAPI(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, session_mock):
        self.session = session_mock
        self.session.api_version = None
        self.session.session_type = None
        self.session.password = "<>"
        self.api = UsersAPI(self.session)
        self.api._endpoints = MagicMock()

    def test_get(self):
        # Arrange
        expected_users = [
            User(
                username="new_user",
                password="new_user",  # pragma: allowlist secret
                group=["netadmin"],
                description="new user",
            )
        ]
        self.api._endpoints.find_users = MagicMock(return_value=expected_users)
        # Act
        observed_users = self.api.get()
        self.api._endpoints.find_users.assert_called_once()
        assert expected_users == observed_users

    def test_get_role(self):
        # Arrange
        expected_user_role = UserRole(isAdmin=True)
        self.api._endpoints.find_user_role = MagicMock(return_value=expected_user_role)
        # Act
        observed_user_role = self.api.get_role()
        # Assert
        self.api._endpoints.find_user_role.assert_called_once()
        assert expected_user_role == observed_user_role

    def test_get_auth_type(self):
        # Arrange
        expected_auth_type = "local"
        self.api._endpoints.find_user_auth_type = MagicMock(
            return_value=UserAuthType(user_auth_type=expected_auth_type)
        )
        # Act
        observed_auth_type = self.api.get_auth_type()
        # Assert
        self.api._endpoints.find_user_auth_type.assert_called_once()
        assert expected_auth_type == observed_auth_type

    def test_create(self):
        # Arrange
        user = User(username="new_user", password="new_user", group=["netadmin"], description="new user")
        self.api._endpoints.create_user = MagicMock()
        # Act
        self.api.create(user)
        # Assert
        self.api._endpoints.create_user.assert_called_once_with(user)

    def test_update(self):
        # Arrange
        user_update = UserUpdateRequest(
            username="new_user",
            password="new_user",  # pragma: allowlist secret
            group=["netadmin"],
            description="new user",
            resource_group="global",
        )
        self.api._endpoints.update_user = MagicMock()
        # Act
        self.api.update(user_update)
        # Assert
        self.api._endpoints.update_user.assert_called_once_with(user_update.username, user_update)

    def test_update_password(self):
        # Arrange
        username = "new_user"
        new_password = "PaSsWoRd"  # pragma: allowlist secret
        self.api._endpoints.update_password = MagicMock()
        # Act
        self.api.update_password(username, new_password)
        # Assert
        self.api._endpoints.update_password.assert_called_once_with(
            username,
            UserUpdateRequest(username=username, password=new_password, current_user_password=self.session.password),
        )

    def test_reset(self):
        # Arrange
        username = "new_user"
        user_reset_request = UserResetRequest(username=username)
        self.api._endpoints.reset_user = MagicMock()
        # Act
        self.api.reset(username)
        # Assert
        self.api._endpoints.reset_user.assert_called_once_with(user_reset_request)

    def test_delete(self):
        # Arrange
        username = "new_user"
        self.api._endpoints.delete_user = MagicMock()
        # Act
        self.api.delete(username)
        # Assert
        self.api._endpoints.delete_user.assert_called_once_with(username)


class TestUserGroupsAPI(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, session_mock):
        self.session = session_mock
        self.session.api_version = None
        self.session.session_type = None
        self.session.password = "<>"
        self.api = UserGroupsAPI(self.session)
        self.api._endpoints = MagicMock()

    def test_get(self):
        # Arrange
        expected_user_groups = [
            UserGroup(
                groupName="new_group", tasks=[UserGroupTask(enabled=True, feature="Alarms", read=True, write=False)]
            )
        ]
        self.api._endpoints.find_user_groups = MagicMock(return_value=expected_user_groups)
        # Act
        observed_user_groups = self.api.get()
        # Assert
        self.api._endpoints.find_user_groups.assert_called_once()
        assert expected_user_groups == observed_user_groups

    def test_create(self):
        # Arrange
        user_group = UserGroup(
            group_name="new_group", tasks=[UserGroupTask(enabled=True, feature="Alarms", read=True, write=False)]
        )
        self.api._endpoints.create_user_group = MagicMock()
        # Act
        self.api.create(user_group)
        # Assert
        self.api._endpoints.create_user_group.assert_called_once_with(user_group)

    def test_update(self):
        # Arrange
        user_group = UserGroup(
            group_name="new_group", tasks=[UserGroupTask(enabled=True, feature="Alarms", read=True, write=False)]
        )
        self.api._endpoints.update_user_group = MagicMock()
        # Act
        self.api.update(user_group)
        # Assert
        self.api._endpoints.update_user_group.assert_called_once_with(user_group.group_name, user_group)

    def test_delete(self):
        # Arrange
        group_name = "new_group"
        self.api._endpoints.delete_user_group = MagicMock()
        # Act
        self.api.delete(group_name)
        # Assert
        self.api._endpoints.delete_user_group.assert_called_once_with(group_name)


class TestResourceGroupsAPI(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, session_mock):
        self.session = session_mock
        self.session.api_version = None
        self.session.session_type = None
        self.session.password = "<>"
        self.api = ResourceGroupsAPI(self.session)
        self.api._endpoints = MagicMock()

    def test_get(self):
        # Arrange
        expected_resource_groups = [
            ResourceGroup(
                id="0:RESGROUP:14567:XD$eD", name="new_resource_group1", desc="New Resource Group #1", site_ids=[200]
            )
        ]
        self.api._endpoints.find_resource_groups = MagicMock(return_value=expected_resource_groups)
        # Act
        observed_resource_groups = self.api.get()
        # Assert
        self.api._endpoints.find_resource_groups.assert_called_once()
        assert expected_resource_groups == observed_resource_groups

    def test_create(self):
        # Arrange
        resource_group = ResourceGroup(name="new_resource_group3", desc="New Resource Group #3", site_ids=[200])
        self.api._endpoints.create_resource_group = MagicMock()
        # Act
        self.api.create(resource_group)
        # Assert
        self.api._endpoints.create_resource_group.assert_called_once_with(resource_group)

    def test_update(self):
        # Arrange
        resource_group_update = ResourceGroupUpdateRequest(
            id="0:RESGROUP:14567:XD$eD", name="new_resource_group1", desc="New Resource Group #1", site_ids=[101, 102]
        )
        self.api._endpoints.update_resource_group = MagicMock()
        # Act
        self.api.update(resource_group_update)
        # Assert
        self.api._endpoints.update_resource_group.assert_called_once_with(
            resource_group_update.id, resource_group_update
        )

    def test_switch(self):
        # Arrange
        resource_group_name = "new_resource_group1"
        self.api._endpoints.switch_resource_group = MagicMock()
        # Act
        self.api.switch(resource_group_name)
        # Assert
        self.api._endpoints.switch_resource_group.assert_called_once_with(
            ResourceGroupSwitchRequest(resource_group_name=resource_group_name)
        )

    def test_delete(self):
        # Arrange
        resource_group_id = "0:RESGROUP:14567:XD$eD"
        self.api._endpoints.delete_resource_group = MagicMock()
        # Act
        self.api.delete(resource_group_id)
        # Assert
        self.api._endpoints.delete_resource_group.assert_called_once_with(resource_group_id)


class TestClusterManagementAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.service_configuration = {
            "vmanageID": "1.1.1.1",
            "deviceIP": "1.1.1.2",
            "services": "none",
        }
        self.service_configuration_dataclass = create_dataclass(ServiceConfigurationData, self.service_configuration)
        self.cluster_management_health_status = [
            {"deviceIP": "localhost"},
            {
                "statistics-db": True,
                "application-server": True,
                "messaging-server": True,
                "configuration-db": True,
                "deviceIP": "localhost",
            },
        ]

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_modify_cluster_setup(self, status_code, expected_outcome, mock_response, mock_session):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = ClusterManagementAPI(mock_session).modify_cluster_setup(self.service_configuration_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.ManagerSession")
    def test_get_cluster_management_health_status(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.cluster_management_health_status
        # Act
        answer = ClusterManagementAPI(mock_session).get_cluster_management_health_status()
        # Assert
        self.assertEqual(answer, self.cluster_management_health_status)


class TestAdministrationSettingsAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.cloud_connector_data = {
            "clientId": "12345",
            "clientSecret": "secret",  # pragma: allowlist secret
            "orgName": "organization",
            "telemetryEnabled": True,
        }
        self.cloud_connector_data_dataclass = create_dataclass(CloudConnectorData, self.cloud_connector_data)
        self.cloud_services_settings = {"enabled": True}
        self.cloud_services_settings_dataclass = create_dataclass(CloudServicesSettings, self.cloud_services_settings)
        self.cloud_on_ramp = [{"": ""}]

    @patch("catalystwan.session.ManagerSession")
    def test_get_sdavc_cloud_connector_config(self, mock_session):
        # Arrange
        mock_session.get_json.return_value = self.cloud_connector_data
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_sdavc_cloud_connector_config()
        # Assert
        self.assertEqual(answer, self.cloud_connector_data_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_enable_sdavc_cloud_connector(self, status_code, expected_outcome, mock_response, mock_session):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).enable_sdavc_cloud_connector(
            self.cloud_connector_data_dataclass
        )
        # Assert
        self.assertEqual(answer, expected_outcome)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_disable_sdavc_cloud_connector(self, status_code, expected_outcome, mock_response, mock_session):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).disable_sdavc_cloud_connector()
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.ManagerSession")
    def test_get_cloud_services(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = [self.cloud_services_settings]
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_cloud_services()
        # Assert
        self.assertEqual(answer, self.cloud_services_settings_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_set_cloud_services(self, status_code, expected_outcome, mock_response, mock_session):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).set_cloud_services(self.cloud_services_settings_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.ManagerSession")
    def test_get_cloud_on_ramp_for_saas_mode(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.cloud_on_ramp
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_cloud_on_ramp_for_saas_mode()
        # Assert
        self.assertEqual(answer, self.cloud_on_ramp)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_enable_cloud_on_ramp_for_saas_mode(self, status_code, expected_outcome, mock_response, mock_session):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).enable_cloud_on_ramp_for_saas_mode(True)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.ManagerSession")
    def test_get_organization(self, mock_session):
        # Arrange
        organization = [{"domain-id": "1", "org": "My org name", "controlConnectionUp": "true"}]
        organization_data = create_dataclass(Organization, organization[0])
        mock_session.get_data.return_value = organization
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_organization()
        # Assert
        self.assertEqual(answer, organization_data)

    @parameterized.expand([[password_dataclass], [certificate_dataclass], [organization_dataclass]])
    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_update(self, payload, mock_response, mock_session):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = 200
        # Act
        answer = AdministrationSettingsAPI(mock_session).update(payload)
        # Assert
        self.assertEqual(answer, True)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_update_vbond(self, mock_response, mock_session):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 200
        # Act
        answer = AdministrationSettingsAPI(mock_session).update(vbond_dataclass)
        # Assert
        self.assertEqual(answer, True)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_update_error_with_dataclass(self, mock_response, mock_session):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 200
        random_dataclass = Connection("random state", "random peer", "127.0.0.1")

        # Act
        def answer():
            AdministrationSettingsAPI(mock_session).update(random_dataclass)

        # Assert
        self.assertRaises(CatalystwanException, answer)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_update_error_without_dataclass(self, mock_response, mock_session):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = 200

        # Act
        def answer():
            AdministrationSettingsAPI(mock_session).update("random_dataclass")

        # Assert
        self.assertRaises(NotAnAttrsClassError, answer)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Optional, cast

from catalystwan.api.task_status_api import Task
from catalystwan.api.versions_utils import DeviceVersions, RepositoryAPI
from catalystwan.endpoints.configuration_device_actions import (
    PartitionActionPayload,
    RemovePartitionActionPayload,
    RemovePartitionDevice,
)
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.exceptions import EmptyVersionPayloadError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.upgrades_helper import get_install_specification, validate_personality_homogeneity

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class PartitionManagerAPI:
    """
    API methods for partitions actions. All methods
    are exececutable on all device categories.

    Usage example:
        # Create session
        session = create_manager_session(...)

        # Prepare devices list
        devices = session.api.devices.get()
        vsmarts = devices.filter(personality=Personality.VSMART)

        # Set current partion as default partition
        partition_manager = PartitionManagerAPI(session)
        set_partition_id = partition_manager.set_default_partition(vsmarts)

        # Check action status
        TaskAPI(session, software_action_id).wait_for_completed()

    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session
        self.repository = RepositoryAPI(self.session)
        self.device_version = DeviceVersions(self.session)

    def set_default_partition(
        self, devices: DataSequence[DeviceDetailsResponse], partition: Optional[str] = None
    ) -> Task:
        """
        Set default software versions for devices

        Args:
            devices (DataSequence[Device]): devices
            partition (Optional[str], optional): If none, current partition will be set as default,
                else selected partition will be set as default

        Returns:
            str: set partition task id
        """
        validate_personality_homogeneity(devices)
        if partition:
            payload_devices = self.device_version.get_device_available(partition, devices)
        else:
            payload_devices = self.device_version.get_devices_current_version(devices)

        for device in payload_devices:
            if not device.version:
                raise EmptyVersionPayloadError("PartitionDevice payload contains entry with empty version field.")

        device_type = get_install_specification(devices.first()).device_type.value
        partition_payload = PartitionActionPayload(
            action="defaultpartition", devices=[dev for dev in payload_devices], device_type=device_type
        )

        partition_action = self.session.endpoints.configuration_device_actions.process_mark_default_partition(
            payload=partition_payload
        )

        return Task(self.session, partition_action.id)

    def remove_partition(
        self, devices: DataSequence[DeviceDetailsResponse], partition: Optional[str] = None, force: bool = False
    ) -> Task:
        """
        Remove chosen software version from device

        Args:
            devices (DataSequence[Device]): remove partition for those devices
            partition (str): If none, all availables partitions will be removed,
                else selected partition will be removed
            force (bool): bypass version checks

        Returns:
            str: action id
        """
        validate_personality_homogeneity(devices)
        if partition:
            payload_devices = self.device_version.get_device_available(partition, devices)
        else:
            payload_devices = self.device_version.get_devices_available_versions(devices)

        for device in payload_devices:
            if not device.version:
                raise EmptyVersionPayloadError("PartitionDevice payload contains entry with empty version field.")

        device_type = get_install_specification(devices.first()).device_type.value
        partition_payload = RemovePartitionActionPayload(
            action="removepartition",
            devices=[RemovePartitionDevice(**dev.model_dump()) for dev in payload_devices],
            device_type=device_type,
        )

        if force is False:
            self._check_remove_partition_possibility(cast(list, partition_payload.devices))

        partition_action = self.session.endpoints.configuration_device_actions.process_remove_partition(
            payload=partition_payload
        )

        return Task(self.session, partition_action.id)

    def _check_remove_partition_possibility(self, payload_devices: List[RemovePartitionDevice]) -> None:
        devices_versions_repository = self.repository.get_devices_versions_repository()
        invalid_devices = []
        for device in payload_devices:
            for version in device.version:
                if version in (
                    devices_versions_repository[device.device_id].current_version,
                    devices_versions_repository[device.device_id].default_version,
                ):
                    invalid_devices.append((device.device_id))
        if invalid_devices:
            raise ValueError(
                f"Current or default version of devices with ids {invalid_devices} \
                    are equal to remove version. Action denied!"
            )



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Optional

from pydantic import AliasPath, BaseModel, Field, model_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import check_fields_exclusive


class GeoLocationListEntry(BaseModel):
    country: Optional[Global[str]] = Field(default=None, description="ISO-3166 alpha-3 country code eg: FRA")
    continent: Optional[Global[str]] = Field(
        default=None, description="One of 2-letter continent codes: AF, NA, OC, AN, AS, EU, SA"
    )

    @model_validator(mode="after")
    def check_country_xor_continent(self):
        check_fields_exclusive(self.__dict__, {"country", "continent"}, True)
        return self


class GeoLocationListParcel(_ParcelBase):
    entries: List[GeoLocationListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_country(self, country: str):
        self.entries.append(GeoLocationListEntry(country=as_global(country)))

    def add_continent(self, continent: str):
        self.entries.append(GeoLocationListEntry(continent=as_global(continent)))



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
import math
from time import sleep
from typing import TYPE_CHECKING, cast
from urllib.error import HTTPError

from catalystwan.api.basic_api import DeviceStateAPI
from catalystwan.dataclasses import Device, Speedtest

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class SpeedtestAPI:
    """
    API methods to perform speedtest between 2 edge devices.

    Example:
        session = create_manager_session(...)
        devices = session.api.basic_api.devices
        speedtest = session.api.speedtest.speedtest(devices[0], devices[1])

        print(f"Upload speed from {speedtest.device_name} is {speedtest.up_speed} Mbps.)

        # example output: Upload speed from vm2 is 217.82 Mbps.

        speedtest is a Speedtest object containing source and destination device info and
            upload and download speed between them
    """

    def __init__(self, session: ManagerSession):
        self.session = session

    def speedtest(
        self,
        source_device: Device,
        destination_device: Device,
        test_duration_seconds: int = 300,
    ) -> Speedtest:
        """
        Performs speedtest between 2 edge devices.

        Args:
            source_device (Device): device from which the speed will be measured
            destination_device (Device): device to which the speed will be measured
            test_duration_seconds (int): duration of the speed measuring in seconds, defaults to 300 (5 minutes)
        """

        source_color = DeviceStateAPI(self.session).get_colors(source_device.id)[0]
        destination_color = DeviceStateAPI(self.session).get_colors(destination_device.id)[0]

        self.speedtest_output = Speedtest(
            device_ip=source_device.local_system_ip,
            device_name=source_device.hostname,
            destination_ip=destination_device.local_system_ip,
            destination_name=destination_device.hostname,
            status="",
            up_speed=0.0,
            down_speed=0.0,
        )  # type: ignore

        if source_device.is_reachable and destination_device.is_reachable:
            with DeviceStateAPI(self.session).enable_data_stream():
                try:
                    self.__perform(
                        source_device,
                        destination_device,
                        source_color,
                        destination_color,
                        test_duration_seconds,
                    )
                except HTTPError as e:
                    self.speedtest_output.status = str(e)
        else:
            self.speedtest_output.status = (
                f"Source is {source_device.reachability.value} and "
                f"destination device is {destination_device.reachability.value}"
            )

        return self.speedtest_output

    def __perform(
        self,
        source_device: Device,
        destination_device: Device,
        source_color: str,
        destination_color: str,
        test_duration_seconds: int = 300,
    ) -> None:
        start_query = {
            "deviceUUID": f"{source_device.uuid}",
            "sourceIp": f"{source_device.local_system_ip}",
            "sourceColor": f"{source_color}",
            "destinationIp": f"{destination_device.local_system_ip}",
            "destinationColor": f"{destination_color}",
            "port": "80",
        }
        url_path = "/dataservice/stream/device/speed"
        setup_speedtest = self.session.post(url_path, json=start_query).json()

        speedtest_session = setup_speedtest["sessionId"]

        self.session.get_json(f"/dataservice/stream/device/speed/start/{speedtest_session}")

        duration = math.ceil(test_duration_seconds / 5)
        for _ in range(duration):
            self.session.get_json(f"/dataservice/stream/device/speed/{speedtest_session}?logId=2")
            sleep(5)

        disable_speedtest = cast(
            dict,
            self.session.get_json(f"/dataservice/stream/device/speed/disable/{speedtest_session}"),
        )

        end_query = {
            "query": {
                "condition": "AND",
                "rules": [
                    {
                        "value": [f"{source_device.local_system_ip}"],
                        "field": "source_local_ip",
                        "type": "string",
                        "operator": "in",
                    },
                    {
                        "value": ["completed"],
                        "field": "status",
                        "type": "string",
                        "operator": "in",
                    },
                ],
            },
            "size": 10000,
        }
        url_path = "/dataservice/statistics/speedtest"
        post_speedtest = self.session.post(url_path, json=end_query).json()

        try:
            self.speedtest_output.status = disable_speedtest["status"]
            self.speedtest_output.up_speed = post_speedtest["data"][0]["up_speed"]
            self.speedtest_output.down_speed = post_speedtest["data"][0]["down_speed"]

            logger.info(
                f"Speedtest from {source_device.local_system_ip} " f"to {destination_device.local_system_ip} succeeded."
            )
        except IndexError:
            self.speedtest_output.status = "No speed received"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import json
import unittest
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

import catalystwan.tests.templates.models as models
from catalystwan.api.template_api import TemplatesAPI
from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.session import ManagerSession

# Take model
# Generate payload
# Compare payload with definition


class TestFeatureTemplate2(TestCase):
    @parameterized.expand(
        [(template,) for template in map(models.__dict__.get, models.__all__)],  # type: ignore
    )
    @patch("catalystwan.session.ManagerSession")
    def test_generate_feature_template_payload_definition(
        self, template: FeatureTemplate, mocked_session: ManagerSession
    ):
        # Arrange
        templates_api = TemplatesAPI(mocked_session)
        with open(Path(__file__).resolve().parents[0] / Path("schemas") / Path(template.type + ".json")) as f:
            schema = json.load(f)

        with open(
            Path(__file__).resolve().parents[0] / Path("definitions") / Path(f"{template.template_name}.json")
        ) as f:
            definition = json.load(f)

        # Act
        feature_template_payload = templates_api.generate_feature_template_payload(
            template=template, schema=schema, debug=False
        )
        # Assert
        # self.assertDictEqual
        self.maxDiff = 10000
        self.assertDictEqual(
            definition["templateDefinition"],
            feature_template_payload.model_dump(by_alias=True, mode="json")["templateDefinition"],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"

from catalystwan.endpoints import APIEndpoints, get
from catalystwan.models.monitoring.server_info import ServerInfoResponse


class ServerInfo(APIEndpoints):
    @get("/server/info")
    def get_server_info(self) -> ServerInfoResponse:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from enum import Enum


class Reachability(str, Enum):
    REACHABLE = "reachable"
    UNREACHABLE = "unreachable"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls, value):
        # Value emitted by Manager ver. 20.10 can be None
        return cls.UNKNOWN



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable


class LocalConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    local: Union[Global[bool], Variable, Default[bool]] = Default[bool](value=False)
    threshold: Optional[Union[Global[int], Variable, Default[None]]] = Default[None](value=None)


class MulticastBasicAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    spt_only: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="sptOnly", validation_alias="sptOnly", default=Default[bool](value=False)
    )
    local_config: LocalConfig = Field(
        serialization_alias="localConfig", validation_alias="localConfig", default=LocalConfig()
    )


class StaticJoin(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_address: Union[Global[str], Variable] = Field(
        serialization_alias="groupAddress", validation_alias="groupAddress"
    )
    source_address: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="sourceAddress", validation_alias="sourceAddress", default=Default[None](value=None)
    )


class IgmpInterfaceParameters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    version: Union[Global[int], Default[int]] = Default[int](value=2)
    join_group: Optional[List[StaticJoin]] = Field(
        serialization_alias="joinGroup", validation_alias="joinGroup", default=None
    )


class IgmpAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface: List[IgmpInterfaceParameters]


class SmmFlag(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable_ssm_flag: Global[bool] = Global[bool](value=True)
    range: Optional[Union[Global[str], Variable, Default[None]]] = Default[None](value=None)


class SptThreshold:
    INFINITY = "infinity"
    ZERO = "0"


class SsmAttrubutes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ssm_range_config: SmmFlag = Field(serialization_alias="ssmRangeConfig", validation_alias="ssmRangeConfig")
    spt_threshold: Optional[Union[Global[SptThreshold], Variable, Default[SptThreshold]]] = Field(
        serialization_alias="sptThreshold",
        validation_alias="sptThreshold",
        default=Default[SptThreshold](value=SptThreshold.ZERO),
    )


class PimInterfaceParameters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    query_interval: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="queryInterval", validation_alias="queryInterval", default=Default[int](value=30)
    )
    join_prune_interval: Union[Global[int], Variable, Default[int]] = Field(
        serialization_alias="joinPruneInterval", validation_alias="joinPruneInterval", default=Default[int](value=60)
    )


class StaticRpAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable]
    access_list: Union[Global[str], Variable] = Field(serialization_alias="accessList", validation_alias="accessList")
    override: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)


class RPAnnounce(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    scope: Union[Global[int], Variable]


class AutoRpAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable_auto_rp_flag: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="enableAutoRPFlag", validation_alias="enableAutoRPFlag", default=Default[bool](value=False)
    )
    send_rp_announce_list: Optional[List[RPAnnounce]] = Field(
        serialization_alias="sendRpAnnounceList", validation_alias="sendRpAnnounceList", default=None
    )
    send_rp_discovery: Optional[List[RPAnnounce]] = Field(
        serialization_alias="sendRpDiscovery", validation_alias="sendRpDiscovery", default=None
    )


class RpDiscoveryScope(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    group_list: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="groupList", validation_alias="groupList", default=None
    )
    interval: Optional[Union[Global[int], Variable, Default[None]]] = None
    priority: Optional[Union[Global[int], Variable, Default[None]]] = None


class BsrCandidateAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(
        serialization_alias="interfaceName", validation_alias="interfaceName"
    )
    mask: Optional[Union[Global[int], Variable, Default[None]]] = None
    priority: Optional[Union[Global[int], Variable, Default[None]]] = None
    accept_rp_candidate: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="acceptRpCandidate", validation_alias="acceptRpCandidate", default=None
    )


class PimBsrAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    rp_candidate: Optional[List[RpDiscoveryScope]] = Field(
        serialization_alias="rpCandidate", validation_alias="rpCandidate", default=None
    )
    bsr_candidate: Optional[List[BsrCandidateAttributes]] = Field(
        serialization_alias="bsdCandidate", validation_alias="bsdCandidate", default=None
    )


class PimAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ssm: SsmAttrubutes
    interface: Optional[List[PimInterfaceParameters]] = None
    rp_addres: Optional[List[StaticRpAddress]] = Field(
        serialization_alias="rpAddr", validation_alias="rpAddr", default=None
    )
    auto_rp: Optional[AutoRpAttributes] = Field(serialization_alias="autoRp", validation_alias="autoRp", default=None)
    pim_bsr: Optional[PimBsrAttributes] = Field(serialization_alias="pimBsr", validation_alias="pimBsr", default=None)


class DefaultMsdpPeer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    default_peer: Union[Global[bool], Default[bool]]
    prefix_list: Optional[Global[UUID]] = None


class MsdpPeerAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    peer_ip: Union[Global[str], Variable] = Field(serialization_alias="peerIp", validation_alias="peerIp")
    connect_source_intf: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="connectSourceIntf", validation_alias="connectSourceIntf", default=None
    )
    remote_as: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="remoteAs", validation_alias="remoteAs", default=None
    )
    password: Optional[Union[Global[str], Variable, Default[None]]] = None
    keepalive_interval: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="keepaliveInterval", validation_alias="keepaliveInterval", default=None
    )
    keepalive_holdtime: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="keepaliveHoldTime", validation_alias="keepaliveHoldTime", default=None
    )
    sa_limit: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="saLimit", validation_alias="saLimit", default=None
    )
    default: Optional[DefaultMsdpPeer] = None


class MsdpPeer(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    mesh_group: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="meshGroup", validation_alias="meshGroup", default=None
    )
    peer: List[MsdpPeerAttributes]


class MsdpAttributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    msdp_list: Optional[List[MsdpPeer]] = Field(
        serialization_alias="msdpList", validation_alias="msdpList", default=None
    )
    originator_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="originatorId", validation_alias="originatorId", default=None
    )
    refresh_timer: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="refreshTimer", validation_alias="refreshTimer", default=None
    )


class MulticastData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    basic: MulticastBasicAttributes = MulticastBasicAttributes()
    igmp: Optional[IgmpAttributes] = None
    pim: Optional[PimAttributes] = None
    msdp: Optional[MsdpAttributes] = None


class MulticastCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: MulticastData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from pathlib import Path
from typing import ClassVar, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.utils.timezone import Timezone


class SystemVsmart(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    timezone: Optional[Timezone] = Field(default=None)
    idle_timeout: Optional[int] = Field(default=None, ge=0, le=300, json_schema_extra={"vmanage_key": "idle-timeout"})
    admin_tech_on_failure: Optional[bool] = Field(
        default=True, json_schema_extra={"vmanage_key": "admin-tech-on-failure"}
    )
    iptables_enable: Optional[bool] = Field(default=True, json_schema_extra={"vmanage_key": "iptables-enable"})
    track_default_gateway: Optional[bool] = Field(
        default=True, json_schema_extra={"vmanage_key": "track-default-gateway"}
    )
    dns_cache_timeout: Optional[int] = Field(
        default=2, ge=1, le=30, json_schema_extra={"vmanage_key": "dns-cache-timeout"}
    )
    track_transport: Optional[bool] = Field(default=True, json_schema_extra={"vmanage_key": "track-transport"})
    controller_group_id: Optional[int] = Field(
        default=0, ge=0, le=100, json_schema_extra={"vmanage_key": "controller-group-id"}
    )
    control_session_pps: Optional[int] = Field(default=300, json_schema_extra={"vmanage_key": "control-session-pps"})
    port_hop: Optional[bool] = Field(default=True, json_schema_extra={"vmanage_key": "port-hop"})
    port_offset: Optional[int] = Field(default=0, ge=0, le=20, json_schema_extra={"vmanage_key": "port-offset"})
    overlay_id: Optional[int] = Field(default=1, ge=1, le=4294967295, json_schema_extra={"vmanage_key": "overlay-id"})
    site_id: Optional[int] = Field(default=1, ge=1, le=4294967295, json_schema_extra={"vmanage_key": "site-id"})
    system_ip: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "system-ip"})
    device_groups: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "device-groups"})
    longitude: Optional[int] = Field(default=None, ge=-180, le=180)
    latitude: Optional[int] = Field(default=None, ge=-90, le=90)
    system_tunnel_mtu: Optional[str] = Field(default=1024, json_schema_extra={"vmanage_key": "system-tunnel-mtu"})
    location: Optional[str] = None
    host_name: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "host-name"})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "system-vsmart"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

from catalystwan.endpoints.configuration_feature_profile import ParcelId, SDRoutingConfigurationFeatureProfile
from catalystwan.models.feature_profile_parcel import FullConfig, FullConfigParcel


class ParcelAPI(Protocol):
    def create(self, name: str, description: str, data: dict) -> ParcelId:
        ...

    def edit(self, parcel_id: str, name: str, description: str, data: dict) -> None:
        ...

    def delete(self, parcel_id: str) -> None:
        ...


class SDRoutingFullConfigParcelAPI(ParcelAPI):
    def __init__(self, session: ManagerSession, fp_id: str):
        self.session = session
        self.fp_id = fp_id
        self.endpoint = SDRoutingConfigurationFeatureProfile(session)

    def create(self, name: str, description: str, data: dict) -> ParcelId:
        payload = FullConfigParcel(name=name, description=description, data=FullConfig(fullconfig=data["fullconfig"]))

        return self.endpoint.create_cli_full_config_parcel(self.fp_id, payload=payload)

    def edit(self, parcel_id: str, name: str, description: str, data: dict) -> None:
        payload = FullConfigParcel(name=name, description=description, data=FullConfig(fullconfig=data["fullconfig"]))

        self.endpoint.edit_cli_full_config_parcel(cli_fp_id=self.fp_id, parcel_id=parcel_id, payload=payload)

    def delete(self, parcel_id: str) -> None:
        self.endpoint.delete_cli_full_config_parcel(cli_fp_id=self.fp_id, parcel_id=parcel_id)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="call-arg"
from unittest import TestCase
from unittest.mock import patch

from catalystwan.api.device_action_api import DecommissionAction, RebootAction, ValidateAction
from catalystwan.dataclasses import Device


class TestRebootActionAPI(TestCase):
    def setUp(self) -> None:
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )

    @patch("catalystwan.session.ManagerSession")
    def test_execute(self, mock_session):
        # Arrange
        reboot_action = RebootAction(mock_session, self.device)
        mock_id = "mock_id"
        mock_session.post.return_value.json.return_value = {"id": mock_id}
        # Act
        reboot_action.execute()
        # Assert
        self.assertEqual(reboot_action.action_id, mock_id)

    @patch("catalystwan.session.ManagerSession")
    def test_execute_raise_exception(self, mock_session):
        # Arrange
        reboot_action = RebootAction(mock_session, self.device)
        mock_session.post.return_value.json.return_value = {}
        # Act&Assert
        self.assertRaises(Exception, reboot_action.execute)


class TestValidateActionAPI(TestCase):
    def setUp(self) -> None:
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )

    @patch("catalystwan.session.ManagerSession")
    def test_execute(self, mock_session):
        # Arrange
        reboot_action = ValidateAction(mock_session, self.device)
        mock_id = "mock_id"
        mock_session.post.return_value.json.return_value = {"id": mock_id}
        # Act
        reboot_action.execute()
        # Assert
        self.assertEqual(reboot_action.action_id, mock_id)

    @patch("catalystwan.session.ManagerSession")
    def test_execute_raise_exception(self, mock_session):
        # Arrange
        reboot_action = ValidateAction(mock_session, self.device)
        mock_session.post.return_value.json.return_value = {}
        # Act&Assert
        self.assertRaises(Exception, reboot_action.execute)


class TestDecommissionActionAPI(TestCase):
    def setUp(self) -> None:
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )

    @patch("requests.Response")
    @patch("catalystwan.session.ManagerSession")
    def test_execute(self, mock_session, mock_response):
        # Arrange
        mock_response.status_code = 200
        reboot_action = DecommissionAction(mock_session, self.device)
        mock_session.put.return_value = mock_response
        # Act
        reboot_action.execute()
        # Assert
        self.assertEqual(mock_response.status_code, 200)

    @patch("requests.Response")
    @patch("catalystwan.session.ManagerSession")
    def test_execute_raise_exception(self, mock_session, mock_response):
        # Arrange
        mock_response.status_code = 404
        # Arrange
        reboot_action = DecommissionAction(mock_session, self.device)
        mock_session.put.return_value = mock_response
        # Act&Assert
        self.assertRaises(Exception, reboot_action.execute)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Optional

from pydantic import BaseModel, Field


class FullConfig(BaseModel):
    fullconfig: str


class FullConfigParcel(BaseModel):
    name: str = Field(pattern=r'^[^&<>! "]+$', min_length=1, max_length=128)
    description: Optional[str]
    data: FullConfig



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Optional

from pydantic import AliasPath, BaseModel, ConfigDict, Field, model_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global
from catalystwan.models.common import TLOCColor
from catalystwan.models.policy.lists_entries import PathPreference


class Preference(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    color_preference: Global[list] = Field(serialization_alias="colorPreference", validation_alias="colorPreference")
    path_preference: Global[PathPreference] = Field(
        serialization_alias="pathPreference", validation_alias="pathPreference"
    )


class PreferredColorGroupEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    primary_preference: Preference = Field(
        serialization_alias="primaryPreference", validation_alias="primaryPreference"
    )
    secondary_preference: Optional[Preference] = Field(
        None, serialization_alias="secondaryPreference", validation_alias="secondaryPreference"
    )
    tertiary_preference: Optional[Preference] = Field(
        None, serialization_alias="tertiaryPreference", validation_alias="tertiaryPreference"
    )

    @model_validator(mode="after")
    def check_passwords_match(self) -> "PreferredColorGroupEntry":
        if not self.secondary_preference and self.tertiary_preference:
            raise ValueError("Preference Entry has to have a secondary prefrence when assigning tertiary preference.")
        return self


class PreferredColorGroupParcel(_ParcelBase):
    entries: List[PreferredColorGroupEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_primary(self, color_preference: List[TLOCColor], path_preference: PathPreference):
        self.entries.append(
            PreferredColorGroupEntry(
                primary_preference=Preference(
                    color_preference=as_global(color_preference),
                    path_preference=as_global(path_preference, PathPreference),
                ),
                secondary_preference=None,
                tertiary_preference=None,
            )
        )

    def add_secondary(self, color_preference: List[TLOCColor], path_preference: PathPreference):
        preferred_color = self.entries[0]
        preferred_color.secondary_preference = Preference(
            color_preference=as_global(color_preference), path_preference=as_global(path_preference, PathPreference)
        )

    def add_tertiary(self, color_preference: List[TLOCColor], path_preference: PathPreference):
        preferred_color = self.entries[0]
        preferred_color.tertiary_preference = Preference(
            color_preference=as_global(color_preference), path_preference=as_global(path_preference, PathPreference)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['catalystwan',
 'catalystwan.api',
 'catalystwan.api.configuration_groups',
 'catalystwan.api.configuration_groups.parcels',
 'catalystwan.api.templates',
 'catalystwan.api.templates.device_template',
 'catalystwan.api.templates.models',
 'catalystwan.api.templates.payloads.aaa',
 'catalystwan.api.templates.payloads.cisco_vpn',
 'catalystwan.api.templates.payloads.tenant',
 'catalystwan.endpoints',
 'catalystwan.endpoints.configuration',
 'catalystwan.endpoints.configuration.device',
 'catalystwan.endpoints.configuration.feature_profile.sdwan',
 'catalystwan.endpoints.configuration.policy',
 'catalystwan.endpoints.configuration.policy.definition',
 'catalystwan.endpoints.configuration.policy.list',
 'catalystwan.endpoints.monitoring',
 'catalystwan.endpoints.real_time_monitoring',
 'catalystwan.endpoints.troubleshooting_tools',
 'catalystwan.integration_tests',
 'catalystwan.models',
 'catalystwan.models.configuration',
 'catalystwan.models.configuration.feature_profile',
 'catalystwan.models.configuration.feature_profile.sdwan.management',
 'catalystwan.models.configuration.feature_profile.sdwan.policy_object',
 'catalystwan.models.configuration.feature_profile.sdwan.policy_object.policy',
 'catalystwan.models.configuration.feature_profile.sdwan.policy_object.security',
 'catalystwan.models.configuration.feature_profile.sdwan.service',
 'catalystwan.models.configuration.feature_profile.sdwan.service.lan',
 'catalystwan.models.configuration.feature_profile.sdwan.system',
 'catalystwan.models.configuration.feature_profile.sdwan.transport',
 'catalystwan.models.misc',
 'catalystwan.models.monitoring',
 'catalystwan.models.policy',
 'catalystwan.models.policy.definitions',
 'catalystwan.tests',
 'catalystwan.tests.templates',
 'catalystwan.tests.templates.definitions',
 'catalystwan.tests.templates.definitions.basic',
 'catalystwan.tests.templates.models',
 'catalystwan.utils',
 'catalystwan.utils.feature_template',
 'catalystwan.workflows']

package_data = \
{'': ['*'],
 'catalystwan.api.templates.payloads.aaa': ['feature/*'],
 'catalystwan.api.templates.payloads.cisco_vpn': ['feature/*'],
 'catalystwan.tests.templates': ['schemas/*', 'schemas/basic/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'attrs>=21.4.0,<22.0.0',
 'ciscoconfparse==1.9.41',
 'clint>=0.5.1,<0.6.0',
 'flake8-quotes>=3.3.1,<4.0.0',
 'packaging>=23.0,<24.0',
 'pydantic>=2.7,<3.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests-toolbelt>=1.0.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'tenacity>=8.1.0,<8.4.0',
 'typing-extensions>=4.6.1,<5.0.0']

setup_kwargs = {
    'name': 'catalystwan',
    'version': '0.33.12rc0',
    'description': 'Cisco Catalyst WAN SDK for Python',
    'long_description': '<p align="center">\n  <a href="#"><img src="docs/images/catalystwan.svg" alt="Cisco Catalyst WAN SDK Logo" style="height:150px" />\n</p>\n\n[![Python-Supported](https://img.shields.io/static/v1?label=Python&logo=Python&color=3776AB&message=3.8%20|%203.9%20|%203.10%20|%203.11%20|%203.12)](https://www.python.org/)\n\nCisco Catalyst WAN SDK is a package for creating simple and parallel automatic requests via official SD-WAN Manager API. It is intended to serve as a multiple session handler (provider, provider as a tenant, tenant). The library is not dependent on environment which is being run in, you just need a connection to any SD-WAN Manager.\n\n## Important Notice: Early Beta Release\n\nWelcome to the Cisco Catalyst WAN SDK!\n\nWe are thrilled to announce that Cisco Catalyst WAN SDK is now available in early beta. This is an exciting step forward in enabling developers to harness the full potential of Cisco\'s networking solutions.  Please be aware that, as an early beta release, this version of the SDK is still undergoing development and testing. As such, it is provided "as is" and support to address any issues are limited and best effort.\n\n## Not recommend to use in production environments.\nWe encourage developers to explore and test the SDK\'s capabilities, but please exercise caution when using it in production environments.  We are dedicated to improving the Cisco Catalyst WAN SDK and we value your input. Your feedback is crucial to us-it will guide us in refining and enhancing the SDK to better meet your needs.\nTo report any issues, share your insights, or suggest improvements, please visit our Issues page on GitHub or reach out to us through the provided communication channels.\n\nThank you for being a part of our development journey!\n\n## Installation\n```console\npip install catalystwan\n```\n\n## Manager Session\nIn order to execute SDK APIs **ManagerSession** needs to be created. The fastest way to get started is to use `create_manager_session()` method which configures session, performs authentication for given credentials and returns **ManagerSession** instance in operational state. **ManagerSession** provides a collection of supported APIs in `api` instance variable.\nPlease check example below:\n\n```python\nfrom catalystwan.session import create_manager_session\n\nurl = "example.com"\nusername = "admin"\npassword = "password123"\n\nwith create_manager_session(url=url, username=username, password=password) as session:\n    devices = session.api.devices.get()\n    print(devices)\n```\n**ManagerSession** extends [requests.Session](https://requests.readthedocs.io/en/latest/user/advanced/#session-objects) so all functionality from [requests](https://requests.readthedocs.io/en/latest/) library is avaiable to user, it also implements python [contextmanager](https://docs.python.org/3.8/library/contextlib.html#contextlib.contextmanager) and automatically frees server resources on exit.\n\n<details>\n    <summary> <b>Configure Manager Session before using</b> <i>(click to expand)</i></summary>\n\nIt is possible to configure **ManagerSession** prior sending any request.\n\n```python\nfrom catalystwan.session import ManagerSession\n\nurl = "example.com"\nusername = "admin"\npassword = "password123"\n\n# configure session using constructor - nothing will be sent to target server yet\nsession = ManagerSession(url=url, username=username, password=password)\n# login and send requests\nsession.login()\nsession.get("/dataservice/device")\nsession.close()\n```\nWhen interacting with the SDWAN Manager API without using a context manager, it\'s important \nto manually execute the `close()` method to release the user session resource.\nEnsure that the `close()` method is called after you have finished using the session to maintain optimal resource management and avoid potential errors.\n\n</details>\n\n<details>\n    <summary> <b>Login as Tenant</b> <i>(click to expand)</i></summary>\n\nTenant domain needs to be provided in url together with Tenant credentials.\n\n```python\nfrom catalystwan.session import create_manager_session\n\nurl = "tenant.example.com"\nusername = "tenant_user"\npassword = "password123"\n\nwith create_manager_session(url=url, username=username, password=password) as session:\n    print(session.session_type)\n```\n\n</details>\n\n<details>\n    <summary> <b>Login as Provider-as-Tenant</b> <i>(click to expand)</i></summary>\n\nTenant `subdomain` needs to be provided as additional argument together with Provider credentials.\n\n```python\nfrom catalystwan.session import create_manager_session\n\nurl = "example.com"\nusername = "provider"\npassword = "password123"\nsubdomain = "tenant.example.com"\n\nwith create_manager_session(url=url, username=username, password=password, subdomain=subdomain) as session:\n    print(session.session_type)\n```\n\n</details>\n\n\n\n## API usage examples\nAll examples below assumes `session` variable contains logged-in [Manager Session](#Manager-Session) instance.\n\n<details>\n    <summary> <b>Get devices</b> <i>(click to expand)</i></summary>\n\n```python\ndevices = session.api.devices.get()\n```\n\n</details>\n\n<details>\n    <summary> <b>Admin Tech</b> <i>(click to expand)</i></summary>\n\n```Python\nadmin_tech_file = session.api.admin_tech.generate("172.16.255.11")\nsession.api.admin_tech.download(admin_tech_file)\nsession.api.admin_tech.delete(admin_tech_file)\n```\n</details>\n\n<details>\n    <summary> <b>Speed test</b> <i>(click to expand)</i></summary>\n\n```python\ndevices = session.api.devices.get()\nspeedtest = session.api.speedtest.speedtest(devices[0], devices[1])\n```\n\n</details>\n\n<details>\n    <summary> <b>Upgrade device</b> <i>(click to expand)</i></summary>\n\n```python\n# Prepare devices list\ncontrollers = session.endpoints.configuration_device_inventory.get_device_details(\'controllers\')\nvsmarts = controllers.filter(personality=Personality.VSMART)\nimage = "viptela-20.7.2-x86_64.tar.gz"\n\n# Upload image\nsession.api.repository.upload_image(image)\n\n# Install software\n\ninstall_task = session.api.software.install(devices=vsmarts, image=image)\n\n# Check action status\ninstall_task.wait_for_completed()\n```\n\n</details>\n\n<details>\n    <summary> <b>Get alarms</b> <i>(click to expand)</i></summary>\nTo get all alarms:\n\n```python\nalarms = session.api.alarms.get()\n```\n\nTo get all not viewed alarms:\n\n```python\nnot_viewed_alarms = session.api.alarms.get().filter(viewed=False)\n```\n\nTo get all alarms from past `n` hours:\n\n```python\nn = 24\nalarms_from_n_hours = session.api.alarms.get(from_time=n)\n```\n\nTo get all critical alarms from past `n` hours:\n\n```python\nfrom catalystwan.utils.alarm_status import Severity\nn = 48\ncritical_alarms = session.api.alarms.get(from_time=n).filter(severity=Severity.CRITICAL)\n```\n\n</details>\n\n<details>\n    <summary> <b>Users</b> <i>(click to expand)</i></summary>\n\n```python\n# Get all users\nsession.api.users.get()\n\n# Create user\nfrom catalystwan.endpoints.administration_user_and_group import User\nnew_user = User(username="new_user", password="new_user", group=["netadmin"], description="new user")\nsession.api.users.create(new_user)\n\n# Update user data\nnew_user_update = UserUpdateRequest(username="new_user", group=["netadmin", "netops"], locale="en_US", description="updated-new_user-description")\nsession.api.users.update(new_user_update)\n\n# Update user password\nsession.api.users.update_password("new_user", "n3W-P4s$w0rd")\n\n# Reset user\nsession.api.users.reset("new_user")\n\n# Delete user\nsession.api.users.delete("new_user")\n\n# Get current user authentication type and role\nsession.api.users.get_auth_type()\nsession.api.users.get_role()\n```\n\n</details>\n\n<details>\n    <summary> <b>User Groups</b> <i>(click to expand)</i></summary>\n\n```python\n# Get all user groups\nsession.api.user_groups.get()\n\n# Create user group\ngroup = UserGroup("new_user_group", [])\ngroup.enable_read({"Audit Log", "Alarms"})\ngroup.enable_read_and_write({"Device Inventory"})\nsession.api.user_groups.create(group)\n\n# Update user group\ngroup.disable({"Alarms"})\nsession.api.user_groups.update(group)\n\n# Delete user group\nsession.api.user_groups.delete(group.group_name)\n```\n\n</details>\n\n</details>\n\n<details>\n    <summary> <b>Sessions</b> <i>(click to expand)</i></summary>\n\n```python\n# Get all active sessions\nactive_sessions = session.api.sessions.get()\n\n# Invalidate sessions for given user\nnew_user_sessions = active_sessions.filter(raw_username="new_user")\nsession.api.sessions.invalidate(new_user_sessions)\n```\n\n</details>\n\n<details>\n    <summary> <b>Resource Groups</b> <i>(click to expand)</i></summary>\n\n```python\n# get resource groups\nsession.api.resource_groups.get()\n\n# create resource group\nnew_resource_group = ResourceGroup(\n    name="new_resource_group",\n    desc="Custom Resource Group #1",\n    siteIds=[]\n)\nsession.api.resource_groups.create(new_resource_group)\n\n# update resource group\nresource_group = session.api.resource_groups.get().filter(name="new_resource_group").single_or_default()\nupdated_resource_group = ResourceGroupUpdateRequest(\n    id=resource_group.id,\n    name=resource_group.name,\n    desc="Custom Resource Group #1 with updated description and site ids",\n    siteIds=[200]\n)\n\n# switch to resource group view\nsession.api.resource_groups.switch("new_resource_group")\n\n# delete resource group\nsession.api.resource_groups.delete(resource_group.id)\n```\n\n</details>\n\n<details>\n    <summary> <b>Tenant management</b> <i>(click to expand)</i></summary>\n\n```python\napi = session.api.tenant_management\n# create tenants\ntenants = [\n    Tenant(\n        name="tenant1",\n        org_name="CiscoDevNet",\n        subdomain="alpha.bravo.net",\n        desc="This is tenant for unit tests",\n        edge_connector_enable=True,\n        edge_connector_system_ip="172.16.255.81",\n        edge_connector_tunnel_interface_name="GigabitEthernet1",\n        wan_edge_forecast=1,\n    )\n]\ncreate_task = api.create(tenants)\ncreate_task.wait_for_completed()\n# list all tenants\ntenants_data = api.get_all()\n# pick tenant from list by name\ntenant = tenants_data.filter(name="tenant1").single_or_default()\n# get selected tenant id\ntenant_id = tenant.tenant_id\n# get vsession id of selected tenant\nvsessionid = api.vsession_id(tenant_id)\n# delete tenant by ids\ndelete_task = api.delete([tenant_id])\ndelete_task.wait_for_completed()\n# others\napi.get_hosting_capacity_on_vsmarts()\napi.get_statuses()\napi.get_vsmart_mapping()\n```\n</details>\n\n<details>\n    <summary> <b>Tenant migration</b> <i>(click to expand)</i></summary>\n\n```python\nfrom pathlib import Path\nfrom catalystwan.session import create_manager_session\nfrom catalystwan.models.tenant import TenantExport\nfrom catalystwan.workflows.tenant_migration import migration_workflow\n\ntenant = TenantExport(\n    name="mango",\n    desc="Mango tenant description",\n    org_name="Provider Org-Mango Inc",\n    subdomain="mango.fruits.com",\n    wan_edge_forecast=100,\n    migration_key="MangoTenantMigrationKey",   # only for SDWAN Manager >= 20.13\n    is_destination_overlay_mt=True,            # only for SDWAN Manager >= 20.13\n)\n\nwith create_manager_session(url="10.0.1.15", username="st-admin", password="") as origin_session, \\\n     create_manager_session(url="10.9.0.16", username="mt-provider-admin", password="") as target_session:\n    migration_workflow(\n        origin_session=origin_session,\n        target_session=target_session,\n        workdir=Path("workdir"),\n        tenant=tenant,\n        validator="10.9.12.26"\n    )\n```\n\n`migration_workflow` performs multi-step migration procedure according to [Migrate Single-Tenant Cisco SD-WAN Overlay to Multitenant Cisco SD-WAN Deployment](https://www.cisco.com/c/en/us/td/docs/routers/sdwan/configuration/system-interface/vedge-20-x/systems-interfaces-book/sdwan-multitenancy.html#concept_sjj_jmm_z4b)\n\n\nSince 20.13 also MT to ST is supported (just provide suitable origin/target sessions, and `is_destination_overlay_mt` parameter)\n\n\nEach step of the `migration_workflow` procedure can be executed independently using api methods: `export_tenant`, `download`, `import_tenant`, `store_token`, `migrate_network`\n\n```python\norigin_api = origin_session.api.tenant_migration_api\ntarget_api = target_session.api.tenant_migration_api\ntenant_file = Path("~/tenant.tar.gz")\ntoken_file = Path("~/tenant-token.txt")\n# export\nexport_task = origin_api.export_tenant(tenant=tenant)\nremote_filename = export_task.wait_for_file()\n# download\norigin_api.download(export_path, remote_filename)\n# import\nimport_task = target_api.import_tenant(export_path, tenant.migration_key)\nimport_task.wait_for_completed()\n# get token\nmigration_id = import_task.import_info.migration_token_query_params.migration_id\ntarget_api.store_token(migration_id, token_path)\n# migrate network\nmigrate_task = origin_api.migrate_network(token_path)\nmigrate_task.wait_for_completed()\n```\n</details>\n\n### Note:\nTo remove `InsecureRequestWarning`, you can include in your scripts (warning is suppressed when `catalystwan_devel` environment variable is set):\n```Python\nimport urllib3\nurllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)\n```\n\n## Catching Exceptions\n```python\ntry:\n    session.api.users.delete("bogus-user-name")\nexcept ManagerHTTPError as error:\n    # Process an error.\n    print(error.response.status_code)\n    print(error.info.code)\n    print(error.info.message)\n    print(error.info.details)\n\n```\n\n## [Supported API endpoints](https://github.com/cisco-open/cisco-catalyst-wan-sdk/blob/main/ENDPOINTS.md)\n\n\n## [Contributing, bug reporting and feature requests](https://github.com/cisco-open/cisco-catalyst-wan-sdk/blob/main/CONTRIBUTING.md)\n\n## Seeking support\n\nYou can contact us by submitting [issues](https://github.com/cisco-open/cisco-catalyst-wan-sdk/issues), or directly via mail on catalystwan@cisco.com.\n',
    'author': 'kagorski',
    'author_email': 'kagorski@cisco.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/cisco-open/cisco-catalyst-wan-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import re
from datetime import datetime
from email.utils import parsedate_to_datetime
from functools import wraps
from pprint import pformat
from typing import Any, Callable, Dict, Optional, Sequence, Type, TypeVar, Union, cast
from urllib.parse import urlparse

from pydantic import BaseModel
from requests import PreparedRequest, Request, Response
from requests.cookies import RequestsCookieJar
from requests.exceptions import JSONDecodeError

from catalystwan import with_proc_info_header
from catalystwan.abstractions import APIEndpointClientResponse
from catalystwan.exceptions import ManagerErrorInfo
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass

T = TypeVar("T")
PRINTABLE_CONTENT = re.compile(r"(text\/.+)|(application\/(json|html|xhtml|xml|x-www-form-urlencoded))", re.IGNORECASE)
SENSITIVE_URL_PATHS = ["/dataservice/settings/configuration/smartaccountcredentials"]


def response_debug(response: Optional[Response], request: Union[Request, PreparedRequest, None]) -> str:
    """Returns human readable string containing Request-Response contents (helpful for debugging).

    Args:
        response: Response object to be debugged (note it contains an PreparedRequest object already)
        request: optional Request object to be debugged

    When response is provided, request argument is ignored and contents of reqest.response will be returned.

    Returns:
        str
    """
    if request is None:
        if response is None:
            return ""
        else:
            _request: Union[Request, PreparedRequest] = response.request
    else:
        _request = request
    debug_dict = {}
    request_debug = {
        "method": _request.method,
        "url": _request.url,
        "headers": dict(_request.headers.items()),
        "body": getattr(_request, "body", None),
        "json": getattr(_request, "json", None),
    }
    if content_type := {k.lower(): v for k, v in _request.headers.items()}.get("content-type"):
        if not re.search(PRINTABLE_CONTENT, content_type):
            del request_debug["body"]
    if urlparse(_request.url).path in SENSITIVE_URL_PATHS:
        del request_debug["body"]
        del request_debug["json"]
    debug_dict["request"] = {k: v for k, v in request_debug.items() if v is not None}
    if response is not None:
        response_debug = {
            "status": response.status_code,
            "reason": response.reason,
            "elapsed-seconds": round(float(response.elapsed.microseconds) / 1000000, 3),
            "headers": dict(response.headers.items()),
        }
        try:
            json = response.json()

            if isinstance(json, dict):
                json.pop("header", None)

            response_debug.update({"json": json})
        except JSONDecodeError:
            if response.encoding is not None:
                if len(response.text) <= 1024:
                    response_debug.update({"text": response.text})
                else:
                    response_debug.update({"text(trimmed)": response.text[:1024]})
            else:
                response_debug.update({"text(cannot convert to string: unknown encoding)": None})
        debug_dict["response"] = response_debug
    return pformat(debug_dict, width=80, sort_dicts=False)


@with_proc_info_header
def response_history_debug(response: Optional[Response], request: Union[Request, PreparedRequest, None]) -> str:
    """Returns human readable string containing Request-Response history contents for given response.

    Args:
        response: Response object to be debugged (note it contains an PreparedRequest object already)
        request: optional Request object to be debugged (considered to be latest request)

    When response is provided, request argument is ignored and contents of reqest.response will be returned.

    Returns:
        str
    """
    if response is None:
        return response_debug(response, request)
    response_debugs = [response_debug(resp, None) for resp in response.history]
    response_debugs += [response_debug(response, request)]
    return "\n".join(response_debugs)


def parse_cookies_to_dict(cookies: str) -> Dict[str, str]:
    """Utility method to parse cookie string into dict"""
    result: Dict[str, str] = {}
    for item in cookies.split(";"):
        item = item.strip()
        if not item:
            continue
        if "=" not in item:
            result[item] = ""
            continue
        name, value = item.split("=", 1)
        result[name] = value
    return result


class JsonPayload:
    def __init__(self, json: Any = None, empty: bool = False):
        self.json = json
        self.empty = empty
        self.data = None
        self.error = None
        self.headers = None
        if isinstance(json, dict):
            self.data = json.get("data", None)
            self.error = json.get("error", None)
            self.headers = json.get("headers", None)


class ManagerResponse(Response, APIEndpointClientResponse):
    """Extends Response object with methods specific to vManage.
    Object is meant to be created from aready received requests.Response"""

    def __init__(self, response: Response):
        self.__dict__.update(response.__dict__)
        self.jsessionid_expired = self._detect_expired_jsessionid()
        try:
            self.payload = JsonPayload(response.json())
        except JSONDecodeError:
            self.payload = JsonPayload(empty=True)

    def _detect_expired_jsessionid(self) -> bool:
        """Determines if server sent expired JSESSIONID"""
        cookies = self._parse_set_cookie_from_headers()
        if (expires := cookies.get("Expires")) and cookies.get("JSESSIONID"):
            # get current server time, when not present use local time
            # local time might be innacurate but "Expires" is usually set to year 1970
            response_date = self.headers.get("date")
            compare_date = parsedate_to_datetime(response_date) if response_date is not None else datetime.now()
            if parsedate_to_datetime(expires) <= compare_date:
                return True
        return False

    def _parse_set_cookie_from_headers(self) -> RequestsCookieJar:
        """Parses "set-cookie" content from response headers"""
        jar = RequestsCookieJar()
        cookies_string = self.headers.get("set-cookie", "")
        jar.update(parse_cookies_to_dict(cookies_string))
        return jar

    def info(self, history: bool = False) -> str:
        """Returns human readable string containing Request-Response contents
        Args:
            history: include response history (eg. redirects)

        Returns:
            str
        """
        if history:
            return response_history_debug(self, None)
        return response_debug(self, None)

    def dataseq(self, cls: Type[T], sourcekey: Optional[str] = "data", validate: bool = True) -> DataSequence[T]:
        """Returns data contents from JSON payload parsed as DataSequence of Dataclass/BaseModel instances
        Args:
            cls: Dataclass/BaseModel subtype (eg. Devices)
            sourcekey: name of the JSON key from response payload to be parsed. If None whole JSON payload will be used

        Returns:
            DataSequence[T] of given type T which is subclassing from Dataclass/BaseModel,
            in case JSON payload was containing a single Object - sequence with one element is returned
        """
        if self.payload.empty:
            return DataSequence(cls, [])

        if sourcekey is None:
            data = self.payload.json
        else:
            data = self.payload.json.get(sourcekey)

        if isinstance(data, Sequence):
            sequence = data
        else:
            sequence = [cast(dict, data)]

        if issubclass(cls, BaseModel):
            if validate:
                return DataSequence(cls, [cls.model_validate(item) for item in sequence])  # type: ignore
            return DataSequence(cls, [cls.model_construct(**item) for item in sequence])  # type: ignore
        return DataSequence(cls, [create_dataclass(cls, item) for item in sequence])

    def dataobj(self, cls: Type[T], sourcekey: Optional[str] = "data", validate: bool = True) -> T:
        """Returns data contents from JSON payload parsed as Dataclass/BaseModel instance
        Args:
            cls: Dataclass/BaseModel subtype (eg. Devices)
            sourcekey: name of the JSON key from response payload to be parsed. If None whole JSON payload will be used

        Returns:
            Object of given type T which is subclassing from Dataclass/BaseModel,

        """
        if sourcekey is None:
            data = self.payload.json
        else:
            data = self.payload.json.get(sourcekey)

        if issubclass(cls, BaseModel):
            if validate:
                return cls.model_validate(data)  # type: ignore[return-value]
            return cls.model_construct(**data)  # type: ignore[return-value]
        return create_dataclass(cls, data)

    def get_error_info(self) -> ManagerErrorInfo:
        """Returns error information from JSON payload"""

        if self.payload.error is None:
            return ManagerErrorInfo(
                message=None,
                details=None,
                code=None,
            )

        return ManagerErrorInfo(**self.payload.error)


def with_vmanage_response(method: Callable[[Any], Response]) -> Callable[[Any], ManagerResponse]:
    @wraps(method)
    def wrapper(*args, **kwargs) -> ManagerResponse:
        return ManagerResponse(method(*args, **kwargs))

    return wrapper



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable


class StaticMacAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    mac_address: Union[Global[str], Variable] = Field(serialization_alias="macaddr", validation_alias="macaddr")
    vlan: Union[Global[int], Variable]
    interface_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="ifName", validation_alias="ifName", default=None
    )


SwitchportMode = Literal[
    "access",
    "trunk",
]

Duplex = Literal[
    "full",
    "half",
]

PortControl = Literal[
    "auto",
    "force-unauthorized",
    "force-authorized",
]

HostMode = Literal[
    "single-host",
    "multi-auth",
    "multi-host",
    "multi-domain",
]

ControlDirection = Literal[
    "both",
    "in",
]


class SwitchportInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Union[Global[str], Variable] = Field(serialization_alias="ifName", validation_alias="ifName")
    mode: Optional[Global[SwitchportMode]] = None
    shutdown: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=True)
    speed: Optional[Union[Global[str], Variable, Default[None]]] = Default[None](value=None)
    duplex: Optional[Union[Global[Duplex], Variable, Default[None]]] = Default[None](value=None)
    switchport_access_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="switchportAccessVlan", validation_alias="switchportAccessVlan", default=None
    )
    switchport_trunk_allowed_vlans: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="switchportTrunkAllowedVlans", validation_alias="switchportTrunkAllowedVlans", default=None
    )
    switchport_trunk_native_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="switchportTrunkNativeVlan", validation_alias="switchportTrunkNativeVlan", default=None
    )
    port_control: Optional[Union[Global[PortControl], Variable, Default[None]]] = Field(
        serialization_alias="portControl", validation_alias="portControl", default=None
    )
    voice_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="voiceVlan", validation_alias="voiceVlan", default=None
    )
    pae_enable: Optional[Union[Global[bool], Variable, Default[None]]] = Field(
        serialization_alias="paeEnable", validation_alias="paeEnable", default=None
    )
    mac_authentication_bypass: Optional[Union[Global[bool], Variable, Default[None]]] = Field(
        serialization_alias="macAuthenticationBypass", validation_alias="macAuthenticationBypass", default=None
    )
    host_mode: Optional[Union[Global[HostMode], Variable, Default[None]]] = Field(
        serialization_alias="hostMode", validation_alias="hostMode", default=None
    )
    enable_periodic_reauth: Optional[Union[Global[bool], Variable, Default[None]]] = Field(
        serialization_alias="enablePeriodicReauth", validation_alias="enablePeriodicReauth", default=None
    )
    inactivity: Optional[Union[Global[int], Variable, Default[None]]] = None
    reauthentication: Optional[Union[Global[int], Variable, Default[int]]] = None
    control_direction: Optional[Union[Global[ControlDirection], Variable, Default[None]]] = Field(
        serialization_alias="controlDirection", validation_alias="controlDirection", default=None
    )
    restricted_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="restrictedVlan", validation_alias="restrictedVlan", default=None
    )
    guest_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="guestVlan", validation_alias="guestVlan", default=None
    )
    critical_vlan: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="criticalVlan", validation_alias="criticalVlan", default=None
    )
    enable_voice: Optional[Union[Global[bool], Variable, Default[None]]] = Field(
        serialization_alias="enableVoice", validation_alias="enableVoice", default=None
    )


class SwitchportData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface: Optional[List[SwitchportInterface]] = None
    age_time: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ageTime", validation_alias="ageTime", default=Default[int](value=300)
    )
    static_mac_address: Optional[List[StaticMacAddress]] = Field(
        serialization_alias="staticMacAddress", validation_alias="staticMacAddress", default=None
    )


class SwitchportCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: SwitchportData
    metadata: Optional[dict] = None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from enum import Enum
from typing import Union

from pydantic import BaseModel, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable, _ParcelBase


class ConfigTypeValue(Enum):
    NON_E_SIM = "non-eSim"


class ControllerConfig(BaseModel):
    id: Union[Variable, Global[str]] = Field(min_length=1, max_length=5, description="Cellular ID")
    slot: Union[Variable, Global[int], Default[int], None] = Field(
        default=None, description="Set primary SIM slot", ge=0, le=1
    )
    maxRetry: Union[Variable, Global[int], Default[None], None] = Field(
        default=None, description="Set SIM failover retries", ge=0, le=65535
    )
    failovertimer: Union[Variable, Global[int], Default[None], None] = Field(
        default=None, description="Set SIM failover timeout in minutes", ge=3, le=7
    )
    autoSim: Union[Variable, Global[bool], Default[None], None] = Field(
        default=None, description="Enable/Disable Firmware Auto Sim"
    )


class CellularControllerParcel(_ParcelBase):
    config_type: Default[ConfigTypeValue] = Field(default=Default(value=ConfigTypeValue.NON_E_SIM), alias="configType")
    controller_config: ControllerConfig = Field(alias="controllerConfig")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import ipaddress
from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator

DEFAULT_OSPF_HELLO_INTERVAL = 10
DEFAULT_OSPF_DEAD_INTERVAL = 40
DEFAULT_OSPF_RETRANSMIT_INTERVAL = 5
DEFAULT_OSPF_INTERFACE_PRIORITY = 1
DEFAULT_OSPF_REFERENCE_BANDWIDTH = 100
DEFAULT_OSPF_EXTERNAL = 110
DEFAULT_OSPF_INTER_AREA = 110
DEFAULT_OSPF_INTRA_AREA = 110
DEFAULT_OSPF_DELAY = 200
DEFAULT_OSPF_INITIAL_HOLD = 1000
DEFAULT_OSPF_MAX_HOLD = 10000


class MetricType(str, Enum):
    TYPE1 = "type1"
    TYPE2 = "type2"


class Protocol(str, Enum):
    STATIC = "static"
    CONNECTED = "connected"
    BGP = "bgp"
    OMP = "omp"
    NAT = "nat"
    EIGRP = "eigrp"


class Redistribute(FeatureTemplateValidator):
    protocol: Protocol
    route_policy: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "route-policy"})
    dia: Optional[BoolStr] = True
    model_config = ConfigDict(populate_by_name=True)


class AdType(str, Enum):
    ADMINISTRATIVE = "administrative"
    ON_STARTUP = "on-startup"


class RouterLsa(FeatureTemplateValidator):
    ad_type: AdType = Field(json_schema_extra={"vmanage_key": "ad-type"})
    time: int
    model_config = ConfigDict(populate_by_name=True)


class Direction(str, Enum):
    IN = "in"


class RoutePolicy(FeatureTemplateValidator):
    direction: Direction
    pol_name: str = Field(json_schema_extra={"vmanage_key": "pol-name"})
    model_config = ConfigDict(populate_by_name=True)


class Network(str, Enum):
    BROADCAST = "broadcast"
    POINT_TO_POINT = "point-to-point"
    NON_BROADCAST = "non-broadcast"
    POINT_TO_MULTIPOINT = "point-to-multipoint"


class Type(str, Enum):
    SIMPLE = "simple"
    MESSAGE_DIGEST = "message-digest"
    NULL = "null"


class Interface(FeatureTemplateValidator):
    name: str
    hello_interval: Optional[int] = Field(
        DEFAULT_OSPF_DEAD_INTERVAL, json_schema_extra={"vmanage_key": "hello-interval"}
    )
    dead_interval: Optional[int] = Field(DEFAULT_OSPF_DEAD_INTERVAL, json_schema_extra={"vmanage_key": "dead-interval"})
    retransmit_interval: Optional[int] = Field(
        DEFAULT_OSPF_RETRANSMIT_INTERVAL, json_schema_extra={"vmanage_key": "retransmit-interval"}
    )
    cost: Optional[int] = None
    priority: Optional[int] = DEFAULT_OSPF_INTERFACE_PRIORITY
    network: Optional[Network] = Network.BROADCAST
    passive_interface: Optional[BoolStr] = Field(default=False, json_schema_extra={"vmanage_key": "passive-interface"})
    type: Optional[Type] = Field(default=None, json_schema_extra={"data_path": ["authentication"]})
    message_digest_key: Optional[int] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "message-digest-key", "data_path": ["authentication", "message-digest"]},
    )
    md5: Optional[str] = Field(default=None, json_schema_extra={"data_path": ["authentication", "message-digest"]})
    model_config = ConfigDict(populate_by_name=True)


class Range(FeatureTemplateValidator):
    address: ipaddress.IPv4Interface
    cost: Optional[int] = None
    no_advertise: Optional[BoolStr] = Field(default=False, json_schema_extra={"vmanage_key": "no-advertise"})
    model_config = ConfigDict(populate_by_name=True)


class Area(FeatureTemplateValidator):
    a_num: int = Field(json_schema_extra={"vmanage_key": "a-num"})
    stub: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["stub"]}
    )
    nssa: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "no-summary", "data_path": ["nssa"]}
    )
    interface: Optional[List[Interface]] = None
    range: Optional[List[Range]] = None
    model_config = ConfigDict(populate_by_name=True)


class CiscoOSPFModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    router_id: Optional[str] = Field(
        default=None, json_schema_extra={"vmanage_key": "router-id", "data_path": ["ospf"]}
    )
    reference_bandwidth: Optional[int] = Field(
        DEFAULT_OSPF_REFERENCE_BANDWIDTH,
        json_schema_extra={"data_path": ["ospf", "auto-cost"], "vmanage_key": "reference-bandwidth"},
    )
    rfc1583: Optional[BoolStr] = Field(default=True, json_schema_extra={"data_path": ["ospf", "compatible"]})
    originate: Optional[BoolStr] = Field(default=None, json_schema_extra={"data_path": ["ospf", "default-information"]})
    always: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"data_path": ["ospf", "default-information", "originate"]}
    )
    metric: Optional[int] = Field(
        default=None, json_schema_extra={"data_path": ["ospf", "default-information", "originate"]}
    )
    metric_type: Optional[MetricType] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "metric-type", "data_path": ["ospf", "default-information", "originate"]},
    )
    external: Optional[int] = Field(DEFAULT_OSPF_EXTERNAL, json_schema_extra={"data_path": ["ospf", "distance"]})
    inter_area: Optional[int] = Field(
        DEFAULT_OSPF_INTER_AREA, json_schema_extra={"data_path": ["ospf", "distance"], "vmanage_key": "inter-area"}
    )
    intra_area: Optional[int] = Field(
        DEFAULT_OSPF_INTRA_AREA, json_schema_extra={"data_path": ["ospf", "distance"], "vmanage_key": "intra-area"}
    )
    delay: Optional[int] = Field(DEFAULT_OSPF_DELAY, json_schema_extra={"data_path": ["ospf", "timers", "spf"]})
    initial_hold: Optional[int] = Field(
        DEFAULT_OSPF_INITIAL_HOLD,
        json_schema_extra={"vmanage_key": "initial-hold", "data_path": ["ospf", "timers", "spf"]},
    )
    max_hold: Optional[int] = Field(
        DEFAULT_OSPF_MAX_HOLD, json_schema_extra={"vmanage_key": "max-hold", "data_path": ["ospf", "timers", "spf"]}
    )
    redistribute: Optional[List[Redistribute]] = Field(
        default=None, json_schema_extra={"vmanage_key": "redistribute", "data_path": ["ospf"]}
    )
    router_lsa: Optional[List[RouterLsa]] = Field(
        default=None, json_schema_extra={"vmanage_key": "router-lsa", "data_path": ["ospf", "max-metric"]}
    )
    route_policy: Optional[List[RoutePolicy]] = Field(
        default=None, json_schema_extra={"vmanage_key": "route-policy", "data_path": ["ospf"]}
    )
    area: Optional[List[Area]] = Field(default=None, json_schema_extra={"vmanage_key": "area", "data_path": ["ospf"]})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_ospf"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Any

from catalystwan.api.templates.models.supported import available_models
from catalystwan.exceptions import TemplateTypeError


def choose_model(type_value: str) -> Any:
    """Chooses correct model based on provided type

    With provided type of feature template searches supported by catalystwan models
    and returns correct for given type of feature template class.

    Args:
        type_value: type of feature template

    Returns:
        model

    Raises:
            TemplateTypeError: Raises when the model is not supported by catalystwan.
    """
    if type_value not in available_models:
        for model in available_models.values():
            if model.type == type_value:  # type: ignore
                return model
        raise TemplateTypeError(f"Feature template type '{type_value}' is not supported.")

    return available_models[type_value]



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import MagicMock, Mock, patch

from catalystwan.api.software_action_api import SoftwareActionAPI
from catalystwan.api.versions_utils import DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from catalystwan.endpoints.configuration.software_actions import SoftwareImageDetails
from catalystwan.endpoints.configuration_device_actions import ActionId, InstallDevice, PartitionDevice
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.exceptions import ImageNotInRepositoryError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.upgrades_helper import Family, InstallSpecHelper


class TestSoftwareAcionAPI(unittest.TestCase):
    def setUp(self):
        self.device = DeviceDetailsResponse(
            personality="vedge",
            uuid="mock_uuid",
            device_ip="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                installed_versions=["ver1", "ver2", "curr_ver"],
                availableVersions=["ver1", "ver2"],
                version="20.8",
                defaultVersion="def_ver",
                uuid="mock_uuid",
            ),
        }

        self.mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]
        self.install_spec = InstallSpecHelper.VMANAGE.value

        mock_session = Mock()
        self.mock_repository_object = RepositoryAPI(mock_session)
        self.mock_device_versions = DeviceVersions(self.mock_repository_object)
        self.mock_software_action_obj = SoftwareActionAPI(mock_session)

    @patch.object(SoftwareActionAPI, "_downgrade_check")
    @patch.object(RepositoryAPI, "get_image_version")
    def test_upgrade_software_if_downgrade_check_is_none(self, mock_get_image_version, mock_downgrade_check):
        # Prepare mock data
        mock_downgrade_check.return_value = False
        expected_id = ActionId(id="mock_action_id")
        mock_get_image_version.return_value = "any_version"
        self.mock_software_action_obj.session.endpoints.configuration_device_actions.process_install_operation = (
            MagicMock(return_value=expected_id)
        )

        # Assert
        answer = self.mock_software_action_obj.install(
            devices=DataSequence(DeviceDetailsResponse, [self.device]),
            reboot=True,
            sync=True,
            image="path",
        )
        self.assertEqual(answer.task_id, "mock_action_id")

    @patch.object(DeviceVersions, "get_device_available")
    @patch.object(RepositoryAPI, "get_all_software_images")
    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_activate_software(
        self, mock_get_devices_versions_repository, mock_get_all_software_images, mock_get_device_available
    ):
        # Prepare mock data
        expected_id = ActionId(id="mock_action_id")
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_get_device_available.return_value = DataSequence(
            PartitionDevice, [PartitionDevice(device_id="mock_uuid", device_ip="mock_ip", version="ver2")]
        )
        mock_get_all_software_images.return_value = DataSequence(
            SoftwareImageDetails,
            [SoftwareImageDetails(**{"availableFiles": "vmanage-20.9.1-x86_64.tar.gz", "versionName": "ver2"})],
        )

        self.mock_software_action_obj.session.endpoints.configuration_device_actions.process_mark_change_partition = (
            MagicMock(return_value=expected_id)
        )

        # Assert
        answer = self.mock_software_action_obj.activate(
            devices=DataSequence(DeviceDetailsResponse, [self.device]),
            image="vmanage-20.9.1-x86_64.tar.gz",
        )
        self.assertEqual(answer.task_id, "mock_action_id")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_downgrade_check_no_incorrect_devices(self, mock_get_devices_versions_repository):
        # Preapre mock data
        upgrade_version = "20.9"
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [InstallDevice(**{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": upgrade_version})]
        self.mock_repository_object.devices = mock_devices
        # Assert
        answer = self.mock_software_action_obj._downgrade_check(mock_devices, upgrade_version, Family.VMANAGE.value)
        self.assertEqual(answer, None, "downgrade detected, but should not")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_downgrade_check_incorrect_devices_exists(self, mock_get_devices_versions_repository):
        # Preapre mock data
        upgrade_version = "20.6"
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [InstallDevice(**{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": upgrade_version})]
        self.mock_repository_object.devices = mock_devices
        # Assert
        # answer = self.mock_software_action_obj._downgrade_check(mock_devices, upgrade_version, Family.VMANAGE.value)
        self.assertRaises(
            ValueError,
            self.mock_software_action_obj._downgrade_check,
            mock_devices,
            upgrade_version,
            Family.VMANAGE.value,
        )

    def test_install_software_from_remote_image_not_available_with_downgrade_check(self):
        with self.assertRaises(ValueError):
            self.mock_software_action_obj.install(
                devices=DataSequence(DeviceDetailsResponse, [self.device]),
                remote_server_name="dummy",
                remote_image_filename="dummy",
            )

    @patch.object(RepositoryAPI, "get_all_software_images")
    def test_install_software_from_remote_image_with_wrong_version(self, mock_get_all_software_images):
        mock_get_all_software_images.return_value = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": "ver1",
                    }
                )
            ],
        )

        with self.assertRaises(ImageNotInRepositoryError):
            self.mock_software_action_obj.install(
                devices=DataSequence(DeviceDetailsResponse, [self.device]),
                remote_server_name="remote-server-test",
                remote_image_filename="not-ver1",
                downgrade_check=False,
            )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum


class PrintColors(Enum):
    RED_BACKGROUND: str = "\033[41m"
    RED: str = "\033[31m"
    YELLOW: str = "\033[33m"
    BLUE: str = "\033[34m"
    GREEN: str = "\033[32m"
    NONE: str = "\033[0m"



# Copyright 2022 Cisco Systems, Inc. and its affiliates

import unittest
from typing import Optional
from unittest.mock import patch
from uuid import uuid4

import pytest  # type: ignore
from parameterized import parameterized  # type: ignore
from requests import HTTPError, Request, RequestException, Response

from catalystwan.exceptions import CatalystwanException, ManagerHTTPError, ManagerRequestException
from catalystwan.session import ManagerSession


@pytest.mark.skip(reason="Session is not mocked property (#149)")
class TestSession(unittest.TestCase):
    def setUp(self):
        self.url = "example.com"
        self.username = "admin"
        self.password = str(uuid4())

    def test_session_str(self):
        # Arrange, Act
        session = ManagerSession(self.url, self.username, self.password, port=111)

        # Assert
        self.assertEqual(str(session), "admin@https://example.com:111")

    @parameterized.expand(
        [
            (None, "http://example.com:666", "http://example.com:666"),
            (None, "www.example.com", "https://www.example.com"),
            (123, "example.com", "https://example.com:123"),
            (123, "http://example.com", "http://example.com:123"),
            (123, "https://example.com", "https://example.com:123"),
            (None, "https://example.com", "https://example.com"),
        ]
    )
    def test_base_url(self, port: Optional[int], user_url: str, expected_url: str):
        # Arrange, Act
        session = ManagerSession(user_url, self.username, self.password, port=port)
        # Assert
        self.assertEqual(session.base_url, expected_url)

    def test_session_repr(self):
        # Arrange, Act
        session = ManagerSession("domain.com", "user1", "$password", port=111)
        session_str = "ManagerSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        # Assert
        self.assertEqual(repr(session), session_str)

    def test_session_repr_different_sessions(self):
        # Arrange, Act
        session = ManagerSession("domain.com", "user1", "$password", port=111)
        session_str = "ManagerSession('not.domain.com', 'different_user', '$password', port=111, subdomain='None')"
        # Assert
        self.assertNotEqual(repr(session), session_str)

    @patch("catalystwan.session.Session.__repr__")
    def test_session_eval_repr(self, mock_repr):
        # Arrange, Act
        mock_repr.return_value = "ManagerSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        session = ManagerSession("domain.com", "user1", "$password", port=111)
        # Assert
        self.assertEqual(eval(mock_repr()), session)

    @patch("catalystwan.session.ManagerSession.check_vmanage_server_connection")
    @patch("catalystwan.session.Session.__repr__")
    def test_session_eval_repr_different_sessions(self, mock_repr, mock_check_connection):
        # Arrange, Act
        mock_check_connection.return_value = True
        mock_repr.return_value = "ManagerSession('domain.com', 'user1', '$password', port=111, subdomain='None')"
        session = ManagerSession("not.domain.com", "different_user", "$password", port=111)
        # Assert
        self.assertNotEqual(eval(mock_repr()), session)

    def test_session_eq(self):
        # Arrange, Act
        session_1 = ManagerSession("domain.com", "user1", "$password", port=111)
        session_2 = ManagerSession("domain.com", "user1", "$password", port=111)
        # Assert
        self.assertEqual(session_1, session_2)

    @patch("catalystwan.session.ManagerSession.check_vmanage_server_connection")
    def test_session_eq_different_sessions(self, mock_check_connection):
        # Arrange, Act
        mock_check_connection.return_value = True
        session_1 = ManagerSession("domain.com", "user1", "$password", port=111)
        session_2 = ManagerSession("not.domain.com", "different_user", "$password", port=111)
        # Assert
        self.assertNotEqual(session_1, session_2)

    @parameterized.expand(
        [
            (123, "/devices", "https://example.com:123/devices"),
            (123, "devices", "https://example.com:123/devices"),
            (None, "/devices", "https://example.com/devices"),
            (None, "devices", "https://example.com/devices"),
        ]
    )
    def test_get_full_url(self, port: Optional[int], url: str, full_url: str):
        # Arrange, Act
        session = ManagerSession(self.url, self.username, self.password, port=port)

        # Assert
        self.assertEqual(session.get_full_url(url), full_url)

    @patch("catalystwan.session.head")
    def test_check_vmanage_server_with_port(self, mock_head):
        # Arrange, Act
        mock_head.return_value = None
        session = ManagerSession("domain.com", "user1", "$password", port=111)
        answer = session.check_vmanage_server_connection()
        # Assert
        self.assertEqual(answer, True)

    @patch("catalystwan.session.head")
    def test_check_vmanage_server_no_port(self, mock_requests):
        # Arrange, Act
        mock_requests.return_value = None
        session = ManagerSession("domain.com", "user1", "$password")
        answer = session.check_vmanage_server_connection()
        # Assert
        self.assertEqual(answer, True)


class TestSessionExceptions(unittest.TestCase):
    def setUp(self):
        self.session = ManagerSession(url="domain.com", username="user", password="<>")
        response = Response()
        response.json = lambda: {
            "error": {
                "message": "Delete users request failed",
                "details": "No user with name XYZ was found",
                "code": "USER0006",
            }
        }
        response.status_code = 500
        response.request = Request(method="GET", url="/v1/data")
        self.response = response

    @parameterized.expand(
        [
            (RequestException(), [CatalystwanException, ManagerRequestException, RequestException]),
            (HTTPError(), [CatalystwanException, ManagerHTTPError, ManagerRequestException, RequestException]),
        ]
    )
    @patch("requests.sessions.Session.request")
    def test_session_request_exceptions(self, lib_exception, sdk_exceptions, mock_request_base):
        # Arrange
        if isinstance(lib_exception, HTTPError):
            mock_request_base.return_value = self.response
        else:
            mock_request_base.side_effect = lib_exception
        for sdk_exception in sdk_exceptions:
            # Act / Assert
            with self.assertRaises(sdk_exception):
                self.session.request(self.response.request.method, self.response.request.url)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import MagicMock, patch

from catalystwan.api.task_status_api import Task
from catalystwan.api.tenant_management_api import TenantManagementAPI
from catalystwan.endpoints.tenant_management import (
    ControlStatus,
    SiteHealth,
    TenantStatus,
    TenantUpdateRequest,
    vEdgeHealth,
    vSessionId,
    vSmartPlacementUpdateRequest,
    vSmartStatus,
    vSmartTenantCapacity,
    vSmartTenantMap,
)
from catalystwan.models.tenant import Tenant
from catalystwan.typed_list import DataSequence


class TenantManagementAPITest(unittest.TestCase):
    @patch("catalystwan.session.ManagerSession")
    def setUp(self, session_mock):
        self.session = session_mock
        self.session.api_version = None
        self.session.session_type = None
        self.api = TenantManagementAPI(self.session)

    def test_get(self):
        expected_tenants = [
            Tenant(
                name="tenant1",
                org_name="CiscoDevNet",
                subdomain="alpha.bravo.net",
                desc="This is tenant for unit tests",
                edge_connector_enable=True,
                edge_connector_system_ip="172.16.255.81",
                edge_connector_tunnel_interface_name="GigabitEthernet1",
                wan_edge_forecast=1,
            )
        ]
        self.api._endpoints.get_all_tenants = MagicMock(return_value=expected_tenants)
        observed_tenants = self.api.get()
        assert expected_tenants == observed_tenants

    def test_create(self):
        tenants = [
            Tenant(
                name="tenant1",
                org_name="CiscoDevNet",
                subdomain="alpha.bravo.net",
                desc="This is tenant for unit tests",
                edge_connector_enable=True,
                edge_connector_system_ip="172.16.255.81",
                edge_connector_tunnel_interface_name="GigabitEthernet1",
                wan_edge_forecast=1,
            )
        ]
        task = self.api.create(tenants)
        self.assertIsInstance(task, Task)

    def test_update(self):
        # Arrange
        tenant_update_request = TenantUpdateRequest(
            tenant_id="apo605#",
            subdomain="doamin.tenant.net",
            desc="Tenant1 description",
            wan_edge_forecast=1,
            edge_connector_enable=False,
        )
        self.api._endpoints.update_tenant = MagicMock()
        # Act
        self.api.update(tenant_update_request=tenant_update_request)
        # Assert
        self.api._endpoints.update_tenant.assert_called_once_with(
            tenant_id=tenant_update_request.tenant_id, tenant_update_request=tenant_update_request
        )

    def test_update_vsmart_placement(self):
        # Arrange
        src_uuid = "123190GDS*!"
        dst_uuid = "!_0ac%$asfDS"
        tenant_id = "apo605#"
        vsmart_placement_update = vSmartPlacementUpdateRequest(
            src_vsmart_uuid=src_uuid,
            dest_vsmart_uuid=dst_uuid,
        )
        self.api._endpoints.update_tenant_vsmart_placement = MagicMock()
        # Act
        self.api.update_vsmart_placement(tenant_id=tenant_id, src_vsmart_uuid=src_uuid, dst_vsmart_uuid=dst_uuid)
        # Assert
        self.api._endpoints.update_tenant_vsmart_placement.assert_called_once_with(
            tenant_id=tenant_id, vsmart_placement_update_request=vsmart_placement_update
        )

    def test_delete(self):
        tenant_id_list = ["1"]
        password = "password"  # pragma: allowlist secret
        task = self.api.delete(tenant_id_list, password)
        self.assertIsInstance(task, Task)

    def test_delete_auto_password(self):
        tenant_id_list = ["1"]
        self.session.password = "p4s$w0rD"  # pragma: allowlist secret
        task = self.api.delete(tenant_id_list)
        self.assertIsInstance(task, Task)

    def test_get_statuses(self):
        tenant_status = TenantStatus(
            tenant_id="tenant2",
            tenant_name="TeanantTwo",
            control_status=ControlStatus(control_up=1, control_down=0, partial=1),
            site_health=SiteHealth(full_connectivity=2, partial_connectivity=1, no_connectivity=0),
            vedge_health=vEdgeHealth(normal=3, warning=1, error=0),
            vsmart_status=vSmartStatus(up=1, down=0),
        )
        expected_statuses = DataSequence(TenantStatus, [tenant_status])
        self.api._endpoints.get_all_tenant_statuses = MagicMock(return_value=expected_statuses)
        observed_statuses = self.api.get_statuses()
        assert expected_statuses == observed_statuses

    def test_get_hosting_capacity_on_vsmarts(self):
        capacity = vSmartTenantCapacity(vsmart_uuid="ABCD-1234", total_tenant_capacity=12, current_tenant_count=5)
        expected_capacities = DataSequence(vSmartTenantCapacity, [capacity])
        self.api._endpoints.get_tenant_hosting_capacity_on_vsmarts = MagicMock(return_value=expected_capacities)
        observed_capacities = self.api.get_hosting_capacity_on_vsmarts()
        assert expected_capacities == observed_capacities

    def test_get_vsmart_mapping(self):
        expected_mapping = vSmartTenantMap(
            data={
                "vsmart1": [
                    Tenant(
                        name="tenant1",
                        org_name="Tenant1-organization",
                        desc="Tenant1 description",
                        subdomain="tenant1.organization.org",
                        flake_id=9987,
                    )
                ]
            }
        )
        self.api._endpoints.get_tenant_vsmart_mapping = MagicMock(return_value=expected_mapping)
        observed_mapping = self.api.get_vsmart_mapping()
        assert expected_mapping == observed_mapping

    def test_vsession_id(self):
        expected_vsession_id = "567-DEF"
        self.api._endpoints.vsession_id = MagicMock(return_value=vSessionId(vsessionid=expected_vsession_id))
        observed_vsession_id = self.api.vsession_id("1")
        assert expected_vsession_id == observed_vsession_id



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from catalystwan.endpoints import APIEndpoints, post
from catalystwan.utils.upgrades_helper import SoftwarePackageUploadPayload


class ConfigurationDeviceSoftwareUpdate(APIEndpoints):
    def download_package_file(self):
        # GET /device/action/software/package/{fileName}
        ...

    def edit_image_metadata(self):
        # PUT /device/action/software/package/{versionId}/metadata
        ...

    def get_image_metadata(self):
        # GET /device/action/software/package/{versionId}/metadata
        ...

    def get_upload_images_count(self):
        # GET /device/action/software/package/imageCount
        ...

    @post("/device/action/software/package")
    def upload_software_to_manager(self, payload: SoftwarePackageUploadPayload) -> None:
        ...

    def process_software_image(self):
        # POST /device/action/software/package/{imageType}
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from uuid import UUID

from catalystwan.endpoints import APIEndpoints, delete, get, post, put, versions
from catalystwan.models.configuration.feature_profile.common import Parcel, ParcelCreationResponse
from catalystwan.models.configuration.feature_profile.sdwan.policy_object import AnyPolicyObjectParcel
from catalystwan.typed_list import DataSequence


class PolicyObjectFeatureProfile(APIEndpoints):
    @versions(supported_versions=(">=20.13"), raises=False)
    @post("/v1/feature-profile/sdwan/policy-object/{profile_id}/{policy_object_list_type}")
    def create(
        self, profile_id: UUID, policy_object_list_type: str, payload: AnyPolicyObjectParcel
    ) -> ParcelCreationResponse:
        ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @post("/v1/feature-profile/sdwan/policy-object/{policy_object_id}/unified/{security_object_list_type}")
    # def create_security_profile_parcel(self, policy_object_id: UUID, security_object_list_type: str):
    #     ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @delete("/v1/feature-profile/sdwan/policy-object/{profile_id}/{policy_object_list_type}/{list_object_id}")
    def delete(self, profile_id: UUID, policy_object_list_type: str, list_object_id: UUID) -> None:
        ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @delete(
    #     "/v1/feature-profile/sdwan/policy-object/{policy_object_id}/unified/{security_object_list_type}/{security_profile_parcel_id}"
    # )
    # def delete_security_profile_parcel1(
    #     self, policy_object_id: UUID, security_object_list_type: str, security_profile_parcel_id: UUID
    # ):
    #     ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @put("/v1/feature-profile/sdwan/policy-object/{profile_id}/{policy_object_list_type}/{list_object_id}")
    def update(
        self, profile_id: UUID, policy_object_list_type: str, list_object_id: UUID, payload: AnyPolicyObjectParcel
    ) -> ParcelCreationResponse:
        ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @put(
    #     "/v1/feature-profile/sdwan/policy-object/{policy_object_id}/unified/{security_object_list_type}/{security_profile_parcel_id}"
    # )
    # def edit_security_profile_parcel1(
    #     self, policy_object_id: UUID, security_object_list_type: str, security_profile_parcel_id: UUID
    # ):
    #     ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @get("/v1/feature-profile/sdwan/policy-object/{profile_id}/{policy_object_list_type}/{list_object_id}")
    def get_by_id(self, profile_id: UUID, policy_object_list_type: str, list_object_id: UUID) -> Parcel:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @get("/v1/feature-profile/sdwan/policy-object/{profile_id}/{policy_object_list_type}", resp_json_key="data")
    def get_all(self, profile_id: UUID, policy_object_list_type: str) -> DataSequence[Parcel]:
        ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @get("/v1/feature-profile/sdwan/policy-object/{policy_object_list_type}/schema")
    # def get_sdwan_policy_object_data_prefix_parcel_schema_by_schema_type(self, policy_object_list_type: str):
    #     ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @get("/v1/feature-profile/sdwan/policy-object/{policy_object_id}/unified/{security_object_list_type}")
    # def get_security_profile_parcel(self, policy_object_id: UUID, security_object_list_type: str):
    #     ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @get(
    #     "/v1/feature-profile/sdwan/policy-object/{policy_object_id}/unified/{security_object_list_type}/{security_profile_parcel_id}"
    # )
    # def get_security_profile_parcel_by_parcel_id(
    #     self, policy_object_id: UUID, security_object_list_type: str, security_profile_parcel_id: UUID
    # ):
    #     ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from pydantic import BaseModel, Field


class ServerInfoResponse(BaseModel):
    """The response may contain an incorrect spelling "Achitecture"."""

    architecture: str = Field(..., alias="Achitecture")
    available_processors: int = Field(
        ..., serialization_alias="Available processors", validation_alias="Available processors"
    )



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Literal
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress

from catalystwan.endpoints import JSON, APIEndpoints, delete, get, post, put
from catalystwan.models.policy.centralized import CentralizedPolicy, CentralizedPolicyEditPayload, CentralizedPolicyInfo
from catalystwan.models.policy.policy import PolicyId
from catalystwan.typed_list import DataSequence


class VSmartConnectivityStatus(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    device_uuid: UUID = Field(serialization_alias="deviceUUID", validation_alias="deviceUUID")
    operation_mode: str = Field(serialization_alias="operationMode", validation_alias="operationMode")
    device_ip: IPvAnyAddress = Field(serialization_alias="deviceIp", validation_alias="deviceIp")
    local_system_ip: IPvAnyAddress = Field(serialization_alias="local-system-ip", validation_alias="local-system-ip")
    is_online: bool = Field(serialization_alias="isOnline", validation_alias="isOnline")


class AutoConfirm(BaseModel):
    confirm: Literal["true"] = "true"


class ActivateDeactivateTaskId(BaseModel):
    id: UUID


class ConfigurationVSmartTemplatePolicy(APIEndpoints):
    @post("/template/policy/vsmart/activate/{id}")
    def activate_policy(
        self, id: UUID, params: AutoConfirm = AutoConfirm(), payload: JSON = {}
    ) -> ActivateDeactivateTaskId:
        ...

    def activate_policy_for_cloud_services(self):
        # POST /template/policy/vsmart/activate/central/{policyId}
        ...

    @get("/template/policy/vsmart/connectivity/status", "data")
    def check_vsmart_connectivity_status(self) -> DataSequence[VSmartConnectivityStatus]:
        ...

    @post("/template/policy/vsmart")
    def create_vsmart_template(self, payload: CentralizedPolicy) -> PolicyId:
        ...

    @post("/template/policy/vsmart/deactivate/{id}")
    def deactivate_policy(
        self, id: UUID, params: AutoConfirm = AutoConfirm(), payload: JSON = {}
    ) -> ActivateDeactivateTaskId:
        ...

    @delete("/template/policy/vsmart/{id}")
    def delete_vsmart_template(self, id: UUID) -> None:
        ...

    @put("/template/policy/vsmart/central/{id}")
    def edit_template_without_lock_checks(self, id: UUID, payload: CentralizedPolicyEditPayload) -> JSON:
        ...

    @put("/template/policy/vsmart/{id}")
    def edit_vsmart_template(self, id: UUID, payload: CentralizedPolicyEditPayload) -> JSON:
        ...

    @get("/template/policy/vsmart", "data")
    def generate_vsmart_policy_template_list(self) -> DataSequence[CentralizedPolicyInfo]:
        ...

    @get("/template/policy/vsmart/definition/{id}")
    def get_template_by_policy_id(self, id: UUID) -> CentralizedPolicy:
        ...

    def qosmos_nbar_migration_warning(self):
        # GET /template/policy/vsmart/qosmos_nbar_migration_warning
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Any, List, Optional

from pydantic import BaseModel, Field

from catalystwan.api.configuration_groups.parcel import Default
from catalystwan.models.configuration.feature_profile.common import (
    DNSIPv4,
    DNSIPv6,
    HostMapping,
    WANIPv4StaticRoute,
    WANIPv6StaticRoute,
)


class ManagementVPN(BaseModel):
    # TODO (mlembke): vpn_id can't have other value, it needs to be constant. How to do that?
    vpn_id: Default[int] = Field(default=Default(value=512), frozen=True, alias="vpnId")
    ipv4_routes: Optional[List[WANIPv4StaticRoute]] = Field(default=[], alias="ipv4Route")
    ipv6_routes: Optional[List[WANIPv6StaticRoute]] = Field(default=[], alias="ipv6Route")
    dns_ipv4: Optional[DNSIPv4] = Field(default=None, alias="dnsIpv4")
    dns_ipv6: Optional[DNSIPv6] = Field(default=None, alias="dnsIpv6")
    new_host_mapping: Optional[List[HostMapping]] = Field(default=[], alias="newHostMapping")
    # TODO (mlembke): add interfaces
    interface: Optional[Any] = Field(default=None)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

ObjectTrackerType = Literal[
    "Interface",
    "SIG",
    "Route",
]


Criteria = Literal[
    "and",
    "or",
]


class SigTracker(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    object_id: Union[Global[int], Variable] = Field(serialization_alias="objectId", validation_alias="objectId")
    object_tracker_type: Global[ObjectTrackerType] = Field(
        serialization_alias="objectTrackerType",
        validation_alias="objectTrackerType",
        default=Global[ObjectTrackerType](value="SIG"),
    )


class InterfaceTracker(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    object_id: Union[Global[int], Variable] = Field(serialization_alias="objectId", validation_alias="objectId")
    object_tracker_type: Global[ObjectTrackerType] = Field(
        serialization_alias="objectTrackerType",
        validation_alias="objectTrackerType",
        default=Global[ObjectTrackerType](value="Interface"),
    )
    interface: Union[Global[str], Variable]


class RouteTracker(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    object_id: Union[Global[int], Variable] = Field(serialization_alias="objectId", validation_alias="objectId")
    object_tracker_type: Global[ObjectTrackerType] = Field(
        serialization_alias="objectTrackerType",
        validation_alias="objectTrackerType",
        default=Global[ObjectTrackerType](value="Route"),
    )
    route_ip: Union[Global[str], Variable] = Field(serialization_alias="routeIp", validation_alias="routeIp")
    route_mask: Union[Global[str], Variable, Default[str]] = Field(
        serialization_alias="routeMask", validation_alias="routeMask"
    )
    vpn: Union[Global[int], Variable, Default[None]] = Default[None](value=None)


class ObjectTrackerCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: Union[InterfaceTracker, RouteTracker, SigTracker]


class ObjectTrackerRef(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tracker_ref: Global[UUID] = Field(serialization_alias="trackerRef", validation_alias="trackerRef")


class ObjectTrackerGroupData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    object_id: Union[Global[int], Variable] = Field(serialization_alias="objectId", validation_alias="objectId")
    tracker_refs: List[ObjectTrackerRef] = Field(serialization_alias="trackerRefs", validation_alias="trackerRefs")
    criteria: Union[Global[Criteria], Variable, Default[Criteria]] = Default[Criteria](value="or")


class ObjectTrackerGroupCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: ObjectTrackerGroupData
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# type: ignore
import datetime
import json
import unittest
from typing import Optional

from attrs import define, field
from parameterized import parameterized

from catalystwan.dataclasses import TLOC, DataclassBase, Device, PacketSetup, TacacsServer, TenantTacacsServer
from catalystwan.utils.creation_tools import FIELD_NAME, asdict, convert_attributes, create_dataclass
from catalystwan.utils.personality import Personality
from catalystwan.utils.reachability import Reachability


@define
class _TestObjectD:
    d: int


@define
class _TestObjectC:
    c: int
    dd: Optional[_TestObjectD] = None


@define
class _TestObjectB:
    rnd: int = field(metadata={FIELD_NAME: "RANDOM"})


@define
class _TestObjectA:
    a: int = field(metadata={FIELD_NAME: "A"})
    b: _TestObjectB = field(metadata={FIELD_NAME: "BBB"})
    c: Optional[_TestObjectC] = None


class TestCreationTools(unittest.TestCase):
    @parameterized.expand(
        [
            (TLOC(1, 1), {"color": 1, "encapsulation": 1}),
            (PacketSetup(1, 2), {"sessionId": 1, "isNewSession": 2}),
            (
                Device(1, "vbond", *range(2), "reachable", 1),
                {
                    "deviceId": 0,
                    "host-name": 1,
                    "local-system-ip": 1,
                    "memState": None,
                    "cpuState": None,
                    "cpuLoad": None,
                    "connectedVManages": [],
                    "device-model": None,
                    "board-serial": None,
                    "vedgeCertificateState": None,
                    "chasisNumber": None,
                    "uuid": 1,
                    "personality": Personality.VBOND,
                    "reachability": Reachability.REACHABLE,
                    "status": None,
                    "memUsage": None,
                    "state_description": None,
                    "site-id": None,
                    "site-name": None,
                },
            ),
            (_TestObjectA(1, _TestObjectB(2)), {"BBB": {"RANDOM": 2}, "A": 1, "c": None}),
            (
                _TestObjectA(1, _TestObjectB(2), _TestObjectC(1)),
                {"BBB": {"RANDOM": 2}, "A": 1, "c": {"c": 1, "dd": None}},
            ),
            (
                _TestObjectA(1, _TestObjectB(2), _TestObjectC(1, _TestObjectD(111))),
                {"BBB": {"RANDOM": 2}, "A": 1, "c": {"c": 1, "dd": {"d": 111}}},
            ),
            (
                TenantTacacsServer(1, "2", [TacacsServer(*range(7)), TacacsServer(*range(2, 9))]),
                {
                    "timeout": 1,
                    "authentication": "2",
                    "server": [
                        {
                            "address": 0,
                            "authPort": 1,
                            "vpn": 2,
                            "vpnIpSubnet": 3,
                            "key": 4,
                            "secretKey": 5,
                            "priority": 6,
                        },
                        {
                            "address": 2,
                            "authPort": 3,
                            "vpn": 4,
                            "vpnIpSubnet": 5,
                            "key": 6,
                            "secretKey": 7,
                            "priority": 8,
                        },
                    ],
                },
            ),
        ]
    )
    def test_asdict(self, obj, serialized_obj):
        # Arrange, Act
        output = asdict(obj)
        print(output)
        # Assert
        self.assertEqual(output, serialized_obj)

    @parameterized.expand(
        [
            (TLOC(1, 1), '{"color": 1, "encapsulation": 1}'),
            (PacketSetup(1, 2), '{"sessionId": 1, "isNewSession": 2}'),
            (
                Device(1, "vbond", *range(2), "reachable", 1, site_id="100", site_name="SITE_100"),
                '{"deviceId": 0, "host-name": 1, "local-system-ip": 1, "memState": null, "cpuState": null, '
                '"cpuLoad": null, "connectedVManages": [], "device-model": null, "board-serial": null, '
                '"vedgeCertificateState": null, "chasisNumber": null, "site-id": "100", "site-name": "SITE_100", '
                '"uuid": 1, "personality": "vbond", "reachability": "reachable", "status": null, "memUsage": null, '
                '"state_description": null}',
            ),
        ]
    )
    def test_asdict_json(self, obj, serialized_obj):
        # Arrange, Act
        output = json.dumps(asdict(obj))
        # Assert
        print(output)
        self.assertEqual(output, serialized_obj)

    def test_convert_attributes(self):
        # Arrange
        @define(field_transformer=convert_attributes)
        class ConvertibleData(DataclassBase):
            date_time_1: datetime.datetime = field(metadata={FIELD_NAME: "fromUnixEpochTimestamp"})
            date_time_2: datetime.datetime = field(metadata={FIELD_NAME: "fromString"})

        conv_data_dict = {
            "fromUnixEpochTimestamp": 1689324694991,
            "fromString": "2023-07-14T07:11:09+00:00",
        }
        # Act
        conv_data = create_dataclass(ConvertibleData, conv_data_dict)
        # Assert
        assert isinstance(conv_data.date_time_1, datetime.datetime)
        assert isinstance(conv_data.date_time_2, datetime.datetime)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import re

from packaging._structures import NegativeInfinity  # type: ignore
from packaging.version import InvalidVersion, Version  # type: ignore


class NullVersion(Version):
    """Represents non-existing `Version` object.

    `NullVersion` instance is considered as the lowest possible version.
    Any other `Version` object will be considered greater than the `NullVersion`.
    """

    def __init__(self):
        """Initialize a NullVersion object."""
        super().__init__("0")
        self._key = (
            NegativeInfinity,
            NegativeInfinity,
            NegativeInfinity,
            NegativeInfinity,
            NegativeInfinity,
            NegativeInfinity,
        )

    def __str__(self) -> str:
        """A representation of the NullVersion for the user.
        >>> NullVersion()
        NullVersion
        """
        return "NullVersion"


def parse_vmanage_version(version: str) -> Version:
    """Parse vmanage version string which could contain custom prefixes and suffixes
    Currently not conforming versions could be found in:
    GET /client/server: '20.12.0-144-li'
    GET /device/action/software/images?imageType=software: 'smart-li-20.13.999-3077'
    Strategy is to iterate by removing potentially problematic part until version is parsed.
    This is temporary solution until custom versioning scheme will be implemented.

    >>> parse_vmanage_version('20.12.0-144-li')
    <Version('20.12.0.post144')>
    >>> parse_vmanage_version('li-20.13.999-3077')
    <Version('20.13.999.post3077')>
    >>> parse_vmanage_version('smart-li-20.13.999-3077')
    <Version('20.13.999.post3077')>
    >>> parse_vmanage_version('Not a version.')
    <class 'catalystwan.version.NullVersion'>

    Args:
        version (str): The version string to parse.

    Returns:
        Version
    """

    def numeric_only(version: str) -> str:
        if match := re.search(r"\d.*\d", version):
            return str(match[0])
        return ""

    for candidate in [version, numeric_only(version)]:
        try:
            return Version(candidate)
        except InvalidVersion:
            continue
    return NullVersion()


def parse_api_version(version: str) -> Version:
    """Parse the given version string.

    When version string is not a valid `Version`, it tries to cut off major and minor part only.
    If that does not work, it produces a `NullVersion` object.

    >>> parse_api_version("20")
    <Version('20.0')>
    >>> parse_api_version("20.12.0-111-xy")
    <Version('20.12')>
    >>> parse_api_version("Not a version.")
    <Version('NullVersion')>

    Args:
        version (str): The version string to parse.

    Returns:
        Version: A Version with supported `major` and `minor` properties.
    """
    parsed_version = parse_vmanage_version(version)
    if not isinstance(parsed_version, NullVersion):
        return Version(f"{parsed_version.major}.{parsed_version.minor}")
    return NullVersion()



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from ipaddress import IPv4Address
from typing import List, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, RootModel

from catalystwan.endpoints import APIEndpoints, get, post
from catalystwan.typed_list import DataSequence


class DisasterRecoveryTaskId(BaseModel):
    id: UUID


class RegisterDisasterRecoveryTaskId(DisasterRecoveryTaskId):
    ...


class DeregisterDisasterRecoveryTaskId(DisasterRecoveryTaskId):
    ...


class ActivateDisasterRecoveryTaskId(DisasterRecoveryTaskId):
    ...


class UnpauseDisasterRecoveryTaskId(DisasterRecoveryTaskId):
    ...


class ReplicationDetailsEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    last_replicated: str = Field(serialization_alias="lastReplicated", validation_alias="lastReplicated")
    export_duration: str = Field(serialization_alias="exportDuration", validation_alias="exportDuration")
    export_size: str = Field(serialization_alias="exportSize", validation_alias="exportSize")
    replication_status: str = Field(serialization_alias="replicationStatus", validation_alias="replicationStatus")
    export_id: str = Field(serialization_alias="exportID", validation_alias="exportID")


class DisasterRecoveryDetailsResponse(BaseModel):
    replication_details: List[ReplicationDetailsEntry]


class ValidateNodePayload(BaseModel):
    ip: IPv4Address
    username: str
    password: str


class ValidatedNodeEntry(BaseModel):
    ip: IPv4Address
    is_reachable: bool


class DisasterRecoveryPauseResponse(BaseModel):
    status: str


class DataCenter(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    name: str
    nms_personality: str = Field(serialization_alias="nmsPersonality", validation_alias="nmsPersonality")
    dc_personality: str = Field(serialization_alias="dcPersonality", validation_alias="dcPersonality")
    management_ip: IPv4Address = Field(serialization_alias="mgmtIPAddress", validation_alias="mgmtIPAddress")
    username: str
    password: str


class DisasterRecoverySettings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    delay_threshold: int = Field(serialization_alias="delayThreshold", validation_alias="delayThreshold")
    start_time: str = Field(serialization_alias="startTime", validation_alias="startTime")
    interval: int


class VbondPayload(BaseModel):
    name: str = Field(default="")
    ip: IPv4Address
    username: str
    password: str


class DisasterRecoveryRegisterPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    data_centers: List[DataCenter] = Field(serialization_alias="dataCenters", validation_alias="dataCenters")
    disaster_recovery_settings: List[DisasterRecoverySettings] = Field(
        serialization_alias="disasterRecoverySettings", validation_alias="disasterRecoverySettings"
    )
    vbonds: List[VbondPayload]


class DisasterRecoveryStatusInformation(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    pause_replication: bool = Field(serialization_alias="pauseReplication", validation_alias="pauseReplication")
    pause_dr: bool = Field(serialization_alias="pauseDR", validation_alias="pauseDR")
    dr_enabled: bool = Field(serialization_alias="drenabled", validation_alias="drenabled")


class DisasterRecoveryStatusResponse(BaseModel):
    dr_status: Union[DisasterRecoveryStatusInformation, None] = Field(
        serialization_alias="drStatus", validation_alias="drStatus"
    )


class DisasterRecoveryScheduleResponse(BaseModel):
    schedule: DisasterRecoverySettings


class RemoteDataCentervManageVersionResponse(RootModel):
    root: list


class RemoteDataCenterDetails(BaseModel):
    name: str
    ip: IPv4Address
    uuid: UUID
    serialno: str
    state: str


class DisasterRecoveryDataCenterStatus(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    dc_personality: str = Field(serialization_alias="dcPersonality", validation_alias="dcPersonality")
    management_ip: IPv4Address = Field(serialization_alias="mgmtIPAddress", validation_alias="mgmtIPAddress")


class LastestLocalHistory(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    last_switch: int = Field(serialization_alias="lastSwitch", validation_alias="lastSwitch")
    reason_for_switch: Union[str, None] = Field(
        serialization_alias="reasonForSwitch", validation_alias="reasonForSwitch"
    )
    updated_primary: Union[str, None] = Field(serialization_alias="updatedPrimary", validation_alias="updatedPrimary")


class ConfigurationDisasterRecovery(APIEndpoints):
    @post("/disasterrecovery/activate")
    def activate(self) -> ActivateDisasterRecoveryTaskId:
        ...

    @post("/disasterrecovery/deregister")
    def delete(self) -> DeregisterDisasterRecoveryTaskId:
        ...

    # def delete_dc(self) -> None:
    #     # POST /disasterrecovery/deleteRemoteDataCenter
    #     ...

    # def delete_local_dc(self) -> None:
    #     # POST /disasterrecovery/deleteLocalDataCenter
    #     ...

    # def disaster_recovery_pause_replication(self) -> None:
    #     # POST /disasterrecovery/pausereplication
    #     ...

    # def disaster_recovery_replication_request(self) -> None:
    #     # POST /disasterrecovery/requestimport
    #     ...

    # def disaster_recovery_un_pause_replication(self) -> None:
    #     # POST /disasterrecovery/unpausereplication
    #     ...

    # def download(self) -> None:
    #     # GET /disasterrecovery/download/backup/{token}/db_bkp.tar.gz
    #     ...

    # def download_replication_data(self) -> None:
    #     # GET /disasterrecovery/download/{token}/{fileName}
    #     ...

    # def get_usernames(self) -> None:
    #     # GET /disasterrecovery/usernames
    #     ...

    # @get("/disasterrecovery/clusterInfo")
    # def get_cluster_info(self) -> None:
    #     ...

    # def get_config_db_restore_status(self) -> None:
    #     # GET /disasterrecovery/dbrestorestatus
    #     ...

    @get("/disasterrecovery/details")
    def get_details(self) -> DisasterRecoveryDetailsResponse:
        ...

    @get("/disasterrecovery/schedule")
    def get_disaster_recovery_local_replication_schedule(self) -> DisasterRecoveryScheduleResponse:
        ...

    @get("/disasterrecovery/drstatus")
    def get_disaster_recovery_status(self) -> DataSequence[DisasterRecoveryDataCenterStatus]:
        ...

    # def get_history(self) -> None:
    #     # GET /disasterrecovery/history
    #     ...

    # def get_local_data_center_state(self) -> None:
    #     # GET /disasterrecovery/localdc
    #     ...

    @get("/disasterrecovery/localLatestHistory")
    def get_local_history(self) -> LastestLocalHistory:
        ...

    @post("/disasterrecovery/validateNodes")
    def get_reachability_info(self, payload: List[ValidateNodePayload]) -> DataSequence[ValidatedNodeEntry]:
        ...

    @get("/disasterrecovery/remotedc")
    def get_remote_data_center_details(self) -> DataSequence[RemoteDataCenterDetails]:
        ...

    @get("/disasterrecovery/remotedc/swversion")
    def get_remote_data_center_version(self) -> RemoteDataCentervManageVersionResponse:
        ...

    # def get_remote_dc_members_state(self) -> None:
    #     # GET /disasterrecovery/remoteDcState
    #     ...

    @get("/disasterrecovery/status")
    def get_dr_status(self) -> DisasterRecoveryStatusResponse:
        ...

    @post("/disasterrecovery/pause")
    def pause_dr(self) -> DisasterRecoveryPauseResponse:
        ...

    # def pause_local_arbitrator(self) -> None:
    #     # POST /disasterrecovery/pauseLocalArbitrator
    #     ...

    # def pause_local_dc_for_dr(self) -> None:
    #     # POST /disasterrecovery/pauseLocalDC
    #     ...

    # def pause_local_dc_replication(self) -> None:
    #     # POST /disasterrecovery/pauseLocalReplication
    #     ...

    @post("/disasterrecovery/register")
    def register(self, payload: DisasterRecoveryRegisterPayload) -> RegisterDisasterRecoveryTaskId:
        ...

    # def restart_data_center(self) -> None:
    #     # POST /disasterrecovery/restartDataCenter
    #     ...

    # def restore_config_db(self) -> None:
    #     # POST /disasterrecovery/dbrestore
    #     ...

    @post("/disasterrecovery/unpause")
    def unpause_dr(self) -> UnpauseDisasterRecoveryTaskId:
        ...

    # def unpause_local_arbitrator(self) -> None:
    #     # POST /disasterrecovery/unpauseLocalArbitrator
    #     ...

    # def unpause_local_dc_for_dr(self) -> None:
    #     # POST /disasterrecovery/unpauseLocalDC
    #     ...

    # def unpause_local_dc_replication(self) -> None:
    #     # POST /disasterrecovery/unpauseLocalReplication
    #     ...

    # def update(self) -> None:
    #     # POST /disasterrecovery/password
    #     ...

    # def update1(self) -> None:
    #     # PUT /disasterrecovery/register
    #     ...

    # def update_disaster_recovery_state(self) -> None:
    #     # POST /disasterrecovery/remotePassword
    #     ...

    # def update_disaster_recovery_state1(self) -> None:
    #     # POST /disasterrecovery/remotedc
    #     ...

    # def update_dr_state(self) -> None:
    #     # POST /disasterrecovery/updateDRConfigOnArbitrator
    #     ...

    # def update_replication(self) -> None:
    #     # POST /disasterrecovery/updateReplication
    #     ...
